#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Reset noise models."""

from typing import Optional

import astropy.constants as const
import numpy as np

from pyxel.detectors import Detector
from pyxel.util import set_random_seed


def compute_ktc_noise(
    temperature: float, capacitance: float, shape: tuple[int, int]
) -> np.ndarray:
    """Compute KTC noise array.

    Parameters
    ----------
    temperature : float
        Temperature. Unit: K
    capacitance : float
        Node capacitance. Unit: F
    shape : tuple
        Shape of the output array.

    Returns
    -------
    np.ndarray
    """

    rms = np.sqrt(const.k_B.value * temperature / capacitance)

    return np.random.normal(0, rms, shape)


def ktc_noise(
    detector: Detector,
    node_capacitance: Optional[float] = None,
    seed: Optional[int] = None,
) -> None:
    """Apply KTC reset noise to detector signal array.

    Parameters
    ----------
    detector : Detector
        Pyxel detector object.
    node_capacitance : float, optional
        Node capacitance. Unit: F
    seed : int, optional
        Random seed.
    """
    with set_random_seed(seed):
        if node_capacitance is not None:
            if node_capacitance <= 0:
                raise ValueError("Node capacitance should be larger than 0!")

            detector.signal.array += compute_ktc_noise(
                temperature=detector.environment.temperature,
                capacitance=node_capacitance,
                shape=detector.geometry.shape,
            )

        else:
            try:
                capacitance = detector.characteristics.node_capacitance

                detector.signal.array += compute_ktc_noise(
                    temperature=detector.environment.temperature,
                    capacitance=capacitance,
                    shape=detector.geometry.shape,
                )

            except AttributeError as ex:
                raise AttributeError(
                    "Characteristic node_capacitance not available for the detector used. "
                    "Please specify node_capacitance in the model argument!"
                ) from ex



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from copy import deepcopy
from pathlib import Path

import pytest

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Detector,
    Environment,
    ReadoutProperties,
)
from pyxel.models import load_detector, save_detector

try:
    # Check if library 'asdf' is installed
    import asdf  # noqa: F401
except ImportError:
    WITH_ASDF = False
else:
    WITH_ASDF = True


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.mark.parametrize(
    "filename",
    [
        pytest.param(Path("detector.h5"), id="hdf5"),
        pytest.param(Path("detector.asdf"), id="asdf"),
    ],
)
def test_save_load_detector(ccd_10x10: CCD, tmp_path: Path, filename: Path):
    """Test methods `save_detector` and `load_detector`."""
    detector: Detector = deepcopy(ccd_10x10)
    exp_detector: Detector = deepcopy(ccd_10x10)

    full_filename: Path = tmp_path / filename
    if full_filename.suffix == ".asdf" and not WITH_ASDF:
        pytest.skip(reason="Missing library 'asdf'")

    # Save to a file
    save_detector(detector=detector, filename=full_filename)
    assert full_filename.exists()

    # Load from a file
    load_detector(detector=detector, filename=full_filename)

    assert detector == exp_detector



"""Create JSON Schema."""

#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import functools
import importlib
import inspect
import textwrap
from collections import defaultdict
from collections.abc import Callable, Iterator, Mapping, Sequence
from dataclasses import dataclass
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Any, Union, get_args, get_origin

import click
from boltons.strutils import under2camel
from numpydoc.docscrape import NumpyDocString
from toolz import dicttoolz
from tqdm.auto import tqdm

from pyxel import __version__
from pyxel.pipelines import DetectionPipeline


@dataclass
class Param:
    description: str
    annotation: Any


@dataclass
class ParamDefault:
    description: str
    annotation: Any
    default: Any


@dataclass
class FuncDocumentation:
    description: str
    parameters: Mapping[str, Param | ParamDefault]


@dataclass
class ModelInfo:
    model_name: str
    model_fullname: str
    model_class_name: str
    func: Callable


@dataclass
class ModelGroupInfo:
    name: str
    class_name: str


@dataclass(frozen=True)
class Klass:
    cls: type
    base_cls: type | None = None

    @property
    def name(self) -> str:
        return self.cls.__name__


def get_annotation(annotation) -> str:
    if get_origin(annotation):
        annotation = str(annotation)
    elif hasattr(annotation, "__name__"):
        annotation = annotation.__name__
    else:
        annotation = str(annotation)

    return annotation


def get_documentation(func: Callable) -> FuncDocumentation:
    assert func.__doc__
    doc = NumpyDocString(inspect.cleandoc(func.__doc__))

    signature: inspect.Signature = inspect.signature(func)

    parameters = {}

    if doc["Parameters"]:
        # Sanity checks
        all_signature_params: set[str] = set(signature.parameters)
        all_doc_params: set[str] = {doc_param.name for doc_param in doc["Parameters"]}

        for doc_param in all_doc_params:
            if ":" in doc_param:
                raise RuntimeError(
                    f"Bad format for {doc_param=}, for function {func=}."
                )

        if not all_doc_params.issubset(all_signature_params):
            missing_params: set[str] = all_doc_params.difference(all_signature_params)
            raise RuntimeError(
                f"Missing key(s) in the signature: {', '.join(missing_params)} for function {func=}."
            )

        for params in doc["Parameters"]:
            name, *_ = params.name.split(":")
            description = "\n".join(params.desc)

            if name not in signature.parameters:
                raise KeyError(
                    f"Missing signature for parameter {name=} in function {func=}."
                )

            parameter: inspect.Parameter = signature.parameters[name]

            annotation: str = get_annotation(parameter.annotation)

            if parameter.default != inspect.Parameter.empty:
                param: ParamDefault | Param = ParamDefault(
                    description=description,
                    annotation=annotation.replace("NoneType", "None"),
                    default=parameter.default,
                )
            else:
                param = Param(description=description, annotation=annotation)

            parameters[name.strip()] = param
    else:
        for name, parameter in signature.parameters.items():
            if parameter.default != inspect.Parameter.empty:
                param: ParamDefault | Param = ParamDefault(
                    description="",
                    annotation=parameter.annotation,
                    default=parameter.default,
                )
            else:
                param = Param(description="", annotation=parameter.annotation)

            parameters[name.strip()] = param

    return FuncDocumentation(
        description="\n".join(doc["Summary"]), parameters=parameters
    )


@functools.cache
def get_doc_from_klass(klass: Klass) -> FuncDocumentation:
    if klass.base_cls is None:
        doc: FuncDocumentation = get_documentation(klass.cls)
    else:
        doc_base: FuncDocumentation = get_documentation(klass.base_cls)
        doc_inherited: FuncDocumentation = get_documentation(klass.cls)

        doc = FuncDocumentation(
            description=doc_inherited.description,
            parameters=dicttoolz.dissoc(doc_inherited.parameters, *doc_base.parameters),
        )

    return doc


def generate_class(klass: Klass) -> Iterator[str]:
    assert isinstance(klass, Klass)

    doc = get_doc_from_klass(klass)
    klass_description_lst: Sequence[str] = textwrap.wrap(doc.description)

    yield "@schema("
    if len(klass_description_lst) == 1:
        yield f"    title={klass.name!r},"
        yield f"    description={klass_description_lst[0]!r}"
    elif len(klass_description_lst) > 1:
        yield f"    title={klass.name!r},"
        yield "    description=("
        for line in klass_description_lst:
            yield f"        {line!r}"
        yield "        )"
    else:
        yield f"    title={klass.name!r}"

    yield ")"
    yield "@dataclass(kw_only=True)  # Python 3.10+"

    if klass.base_cls is None:
        yield f"class {klass.cls.__name__}:"
    else:
        yield f"class {klass.cls.__name__}({klass.base_cls.__name__}):"

    if doc.parameters:
        for name, param in doc.parameters.items():
            title = name

            if (origin := get_origin(param.annotation)) is not None:
                args: Sequence = get_args(param.annotation)

                if origin == Union:
                    if len(args) != 2:
                        raise NotImplementedError

                    annotation: str = f"Optional[{args[0].__name__}]"
                else:
                    raise NotImplementedError
            else:
                annotation = str(param.annotation)

            annotation = annotation.replace("typing.", "")

            yield f"    {name}: {annotation} = field("
            if isinstance(param, ParamDefault):
                yield f"        default={param.default!r},"
            yield "        metadata=schema("

            description_lst: Sequence[str] = textwrap.wrap(param.description)
            if len(description_lst) == 1:
                yield f"            title={title!r},"
                yield f"            description={description_lst[0]!r}"
            elif len(description_lst) > 1:
                yield f"            title={title!r},"
                yield "            description=("
                for line in description_lst:
                    yield f"                    {line!r}"
                yield "                )"
            else:
                yield f"            title={title!r}"

            yield "        )"
            yield "    )"

    else:
        yield "    pass"

    yield ""
    yield ""


def generate_model(
    func: Callable,
    func_name: str,
    func_fullname: str,
    model_name: str,
) -> Iterator[str]:
    doc: FuncDocumentation = get_documentation(func)

    yield "@schema(title='Parameters')"
    yield "@dataclass"
    yield f"class {model_name}Arguments(Mapping[str, Any]):"

    dct = {key: value for key, value in doc.parameters.items() if key != "detector"}

    all_defaults: bool = all([isinstance(el, ParamDefault) for el in dct.values()])

    if dct:
        for name, param in dct.items():
            title = name

            annotation = str(param.annotation).replace("typing.", "")

            yield f"    {name}: {annotation} = field("
            if isinstance(param, ParamDefault):
                yield f"        default={param.default!r},"
            yield "        metadata=schema("
            yield f"            title={title!r}"

            description_lst: Sequence[str] = textwrap.wrap(param.description)
            if len(description_lst) == 1:
                yield f"            ,description={description_lst[0]!r}"
            elif len(description_lst) > 1:
                yield "            ,description=("
                for line in description_lst:
                    yield f"                    {line!r}"
                yield "                )"

            yield "        )"
            yield "    )"
    # else:
    #     yield "    pass"

    yield "    def __iter__(self) -> Iterator[str]:"
    yield f"        return iter({tuple(dct)!r})"

    yield "    def __getitem__(self, item: Any) -> Any:"
    yield "        if item in tuple(self):"
    yield "            return getattr(self, item)"
    yield "        else:"
    yield "            raise KeyError"

    yield "    def __len__(self) -> int:"
    yield f"        return {len(dct)}"

    yield ""
    yield ""
    yield "@schema("
    yield f"    title=\"Model '{func_name}'\""

    description_lst = textwrap.wrap(doc.description)
    if len(description_lst) == 1:
        yield f"    ,description={description_lst[0]!r}"
    elif len(description_lst) > 1:
        yield "    ,description=("
        for line in description_lst:
            yield f"        {line!r}"
        yield "    )"

    yield ")"
    yield "@dataclass"
    yield f"class {model_name}:"
    yield "    name: str"

    if all_defaults:
        yield f"    arguments: {model_name}Arguments = field(default_factory={model_name}Arguments)"
    else:
        yield f"    arguments: {model_name}Arguments"
    yield f"    func: Literal[{func_fullname!r}] = {func_fullname!r}"
    yield "    enabled: bool = True"
    yield ""
    yield ""


def get_model_info(group_name: str) -> Sequence[ModelInfo]:
    group_module = importlib.import_module(f"pyxel.models.{group_name}")

    lst: list[ModelInfo] = []
    for name in dir(group_module):
        if name.startswith("__"):
            continue

        func: Callable = getattr(group_module, name)
        if not callable(func):
            continue

        sig: inspect.Signature = inspect.signature(func)

        if "detector" not in sig.parameters:
            continue

        lst.append(
            ModelInfo(
                model_name=name,
                model_fullname=f"pyxel.models.{group_name}.{name}",
                model_class_name=under2camel(f"Model_{group_name}_{name}"),
                func=func,
            )
        )

    return lst


def capitalize_title(name: str) -> str:
    return " ".join([el.capitalize() for el in name.split("_")])


def generate_group(model_groups_info: Sequence[ModelGroupInfo]) -> Iterator[str]:
    for group_info in model_groups_info:
        group_name = group_info.name

        models_info: Sequence[ModelInfo] = get_model_info(group_name)

        info: ModelInfo
        for info in models_info:
            group_name_title = capitalize_title(group_name)
            model_title = capitalize_title(info.model_name)

            yield "#"
            yield f"# Model: {group_name_title} / {model_title}"
            yield "#"

            yield from generate_model(
                func=info.func,
                func_name=info.model_name,
                func_fullname=info.model_fullname,
                model_name=info.model_class_name,
            )

    yield "#"
    yield "# Detection pipeline"
    yield "#"
    yield "@dataclass"
    yield "class DetailedDetectionPipeline(DetectionPipeline):"

    for group_info in model_groups_info:
        group_name = group_info.name
        group_name_title = capitalize_title(group_name)

        models_info = get_model_info(group_name)
        models_class_names = [info.model_class_name for info in models_info]

        all_model_class_names = ", ".join(
            [
                *models_class_names,
                "ModelLoadDetector",
                "ModelSaveDetector",
                "ModelFunction",
            ]
        )

        yield f"    {group_name}: Optional["
        yield f"        Sequence[Union[{all_model_class_names}]]"
        yield f"    ] = field(default=None, metadata=schema(title={group_name_title!r}))"


def get_model_group_info() -> Sequence[ModelGroupInfo]:
    all_group_models: tuple[str, ...] = DetectionPipeline.MODEL_GROUPS

    lst = [
        ModelGroupInfo(
            name=group_name,
            class_name=under2camel(f"model_group_{group_name}"),
        )
        for group_name in all_group_models
    ]

    return lst


@functools.cache
def create_klass(cls: type | str) -> Klass:
    import pyxel.detectors

    if isinstance(cls, str):
        cls_type: type = getattr(pyxel.detectors, cls)
        return create_klass(cls_type)

    # Try to find a base class
    if (origin := get_origin(cls)) is not None:
        args: Sequence = get_args(cls)

        if origin == Union:
            if len(args) != 2:
                raise NotImplementedError

            # Optional type
            klass = args[0]

        else:
            raise NotImplementedError
    else:
        klass = cls

    _, *base_classes, _ = inspect.getmro(klass)

    if base_classes:
        return Klass(klass, base_cls=base_classes[0])

    return Klass(klass)


def create_graph(cls: type, graph: Mapping[Klass, set[Klass]]) -> None:
    klass: Klass = create_klass(cls)

    doc: FuncDocumentation = get_doc_from_klass(klass)

    if klass.base_cls is None:
        parameters: Mapping[str, Param | ParamDefault] = doc.parameters
    else:
        create_graph(cls=klass.base_cls, graph=graph)
        klass_base: Klass = create_klass(klass.base_cls)
        graph[klass].add(klass_base)

        klass_base_doc: FuncDocumentation = get_doc_from_klass(klass_base)

        parameters = {**doc.parameters, **klass_base_doc.parameters}

    for name, parameter in parameters.items():
        assert parameter.annotation

        klass_param: Klass = create_klass(parameter.annotation)
        graph[klass].add(klass_param)

        if klass_param.base_cls is not None:
            klass_param_base: Klass = create_klass(klass_param.base_cls)
            graph[klass_param].add(klass_param_base)


def generate_detectors() -> Iterator[str]:
    from pyxel.detectors import APD, CCD, CMOS, MKID, Detector

    registered_detectors: Sequence[type[Detector]] = (CCD, CMOS, MKID, APD)

    # Build a dependency graph
    graph: Mapping[Klass, set[Klass]] = defaultdict(set)

    detector: type[Detector]
    for detector in registered_detectors:
        create_graph(cls=detector, graph=graph)

    # Generate code based on the dependency graph
    ts = TopologicalSorter(graph)
    for klass in ts.static_order():
        yield from generate_class(klass)

    yield ""
    yield "#"
    yield "# Outputs"
    yield "#"
    yield "ValidName = Literal["
    yield "    'detector.image.array', 'detector.signal.array', 'detector.pixel.array'"
    yield "]"
    yield "ValidFormat = Literal['fits', 'hdf', 'npy', 'txt', 'csv', 'png', 'jpg', 'jpeg']"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class Outputs:"
    yield "    output_folder: pathlib.Path"
    yield "    save_data_to_file: Optional["
    yield "        Sequence[Mapping[ValidName, Sequence[ValidFormat]]]"
    yield "    ] = None"
    yield ""
    yield ""
    yield "#"
    yield "# Exposure"
    yield "#"
    yield "@dataclass"
    yield "class ExposureOutputs(Outputs):"
    yield "    save_exposure_data: Optional["
    yield "        Sequence[Mapping[str, Sequence[str]]]"
    yield "    ] = None"

    yield "@dataclass"
    yield "class Readout:"
    yield "    times: Union[Sequence, str, None] = None"
    yield "    times_from_file: Optional[str] = None"
    yield "    start_time: float = 0.0"
    yield "    non_destructive: bool = False"
    yield ""
    yield ""
    yield "@schema(title='Exposure')"
    yield "@dataclass"
    yield "class Exposure:"
    yield "    outputs: ExposureOutputs"
    yield "    readout: Readout = field(default_factory=Readout)"
    yield "    result_type: Literal['image', 'signal', 'pixel', 'all'] = 'all'"
    yield "    pipeline_seed: Optional[int] = None"
    yield ""
    yield ""
    yield "#"
    yield "# Observation"
    yield "#"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class ObservationOutputs(Outputs):"
    yield "    save_observation_data: Optional["
    yield "        Sequence[Mapping[str, Sequence[str]]]"
    yield "    ] = None"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class ParameterValues:"
    yield "    key: str"
    yield "    values: Union["
    yield "        Literal['_'],"
    yield "        Sequence[Literal['_']],"
    yield "        Sequence[Union[int, float]],"
    yield "        Sequence[Sequence[Union[int, float]]],"
    yield "        Sequence[Sequence[Sequence[Union[int, float]]]],"
    yield "        Sequence[str],"
    yield "        str, # e.g. 'numpy.unique(...)'"
    yield "    ]"
    yield "    boundaries: Union[Tuple[float, float], Sequence[Tuple[float, float]], None] = None"
    yield "    enabled: bool = True"
    yield "    logarithmic: bool = False"
    yield ""
    yield ""
    yield "@schema(title='Observation')"
    yield "@dataclass"
    yield "class Observation:"
    yield "    outputs: ObservationOutputs"
    yield "    parameters: Sequence[ParameterValues]"
    yield "    readout: Optional[Readout] = None"
    yield "    mode: Literal['product', 'sequential', 'custom'] = 'product'"
    yield "    from_file: Optional[str] = None"
    yield "    column_range: Optional[Tuple[int, int]] = None"
    yield "    with_dask: bool = False"
    yield "    result_type: Literal['image', 'signal', 'pixel', 'all'] = 'all'"
    yield "    pipeline_seed: Optional[int] = None"
    yield ""
    yield ""
    yield "#"
    yield "# Calibration"
    yield "#"
    yield "@dataclass"
    yield "class CalibrationOutputs(Outputs):"
    yield "    save_calibration_data: Optional["
    yield "        Sequence[Mapping[str, Sequence[str]]]"
    yield "    ] = None"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class Algorithm:"
    yield "    type: Literal['sade', 'sga', 'nlopt'] = 'sade'"
    yield "    generations: int = 1"
    yield "    population_size: int = 1"

    yield "    # SADE #####"
    yield "    variant: int = 2"
    yield "    variant_adptv: int = 1"
    yield "    ftol: float = 1e-6"
    yield "    xtol: float = 1e-6"
    yield "    memory: bool = False"

    yield "    # SGA #####"
    yield "    cr: float = 0.9"
    yield "    eta_c: float = 1.0"
    yield "    m: float = 0.02"
    yield "    param_m: float = 1.0"
    yield "    param_s: int = 2"
    yield "    crossover: Literal['single', 'exponential', 'binominal', 'sbx'] = 'exponential'"
    yield "    mutation: Literal['uniform', 'gaussian', 'polynomial'] = 'polynomial'"
    yield "    selection: Literal['tournament', 'truncated'] = 'tournament'"

    yield "    # NLOPT #####"
    yield "    nlopt_solver: Literal["
    yield "        'cobyla', 'bobyqa', 'newuoa', 'newuoa_bound', 'praxis', 'neldermead',"
    yield "        'sbplx', 'mma', 'ccsaq', 'slsqp', 'lbfgs', 'tnewton_precond_restart',"
    yield "        'tnewton_precond', 'tnewton_restart', 'tnewton', 'var2', 'var1', 'auglag',"
    yield "        'auglag_eq'"
    yield "    ] = 'neldermead'"
    yield "    maxtime: int = 0"
    yield "    maxeval: int = 0"
    yield "    xtol_rel: float = 1.0e-8"
    yield "    xtol_abs: float = 0.0"
    yield "    ftol_rel: float = 0.0"
    yield "    ftol_abs: float = 0.0"
    yield "    stopval: Optional[float] = None"
    yield "    # local_optimizer: Optional['pg.nlopt'] = None"
    yield "    replacement: Literal['best', 'worst', 'random'] = 'best'"
    yield "    nlopt_selection: Literal['best', 'worst', 'random'] = 'best'"
    yield ""
    yield ""
    yield "@schema(title='Fitness function')"
    yield "@dataclass"
    yield "class FitnessFunction:"
    yield "    func: str"
    yield "    arguments: Optional[Mapping[str, Any]] = None"
    yield ""
    yield ""
    yield "@schema(title='Calibration')"
    yield "@dataclass"
    yield "class Calibration:"
    yield "    outputs: CalibrationOutputs"
    yield "    target_data_path: Sequence[pathlib.Path]"
    yield "    fitness_function: FitnessFunction"
    yield "    algorithm: Algorithm"
    yield "    parameters: Sequence[ParameterValues]"
    yield "    readout: Optional[Readout] = None"
    yield "    mode: Literal['pipeline', 'single_model'] = 'pipeline'"
    yield "    result_type: Literal['image', 'signal', 'pixel'] = 'image'"
    yield "    result_fit_range: Optional[Sequence[int]] = None"
    yield "    result_input_arguments: Optional[Sequence[ParameterValues]] = None"
    yield "    target_fit_range: Optional[Sequence[int]] = None"
    yield "    pygmo_seed: Optional[int] = None"
    yield "    pipeline_seed: Optional[int] = None"
    yield "    num_islands: int = 1"
    yield "    num_evolutions: int = 1"
    yield "    num_best_decisions: Optional[int] = None"
    yield "    topology: Literal['unconnected', 'ring', 'fully_connected'] = 'unconnected'"
    yield "    type_islands: Literal["
    yield "        'multiprocessing', 'multithreading', 'ipyparallel'"
    yield "    ] = 'multiprocessing'"
    yield "    weights_from_file: Optional[Sequence[pathlib.Path]] = None"
    yield "    weights: Optional[Sequence[float]] = None"

    yield ""
    yield ""
    yield ""

    # Create wrappers for the modes
    mode_classes: Sequence[str] = ("Exposure", "Observation", "Calibration")

    all_configurations: list[str] = []
    for running_mode in mode_classes:
        for detector_klass in registered_detectors:
            detector_name: str = detector_klass.__name__
            klass_name: str = f"Configuration_{running_mode}_{detector_name}"

            yield "@dataclass"
            yield f"class {klass_name}:"
            yield "    pipeline: DetailedDetectionPipeline = field(metadata=schema(title='Pipeline'))"
            yield f"    {running_mode.lower()}: {running_mode} = field(metadata=schema(title={running_mode!r}))"
            yield f"    {detector_name.lower()}_detector: {detector_name} = field(metadata=schema(title={detector_name!r}))"

            all_configurations.append(klass_name)

    return all_configurations


def generate_all_models() -> Iterator[str]:
    lst = get_model_group_info()
    yield "#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022."
    yield "#"
    yield "#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which"
    yield "#  is part of this Pyxel package. No part of the package, including"
    yield "#  this file, may be copied, modified, propagated, or distributed except according to"
    yield "#  the terms contained in the file ‘LICENCE.txt’."
    yield ""
    yield "######################################"
    yield "# Note: This code is auto-generated. #"
    yield "#       Don't modify it !            #"
    yield "######################################"
    yield ""

    yield "import collections"
    yield "import json"
    yield "import pathlib"
    yield "import sys"
    yield "from dataclasses import dataclass, field"
    yield "from pathlib import Path"
    yield "from pprint import pprint"
    yield "from typing import Any, Iterator, Literal, Mapping, Optional, Sequence, Tuple, Union"
    yield ""
    yield "import click"
    yield "from apischema import schema"
    yield "from apischema.json_schema import JsonSchemaVersion, deserialization_schema"

    yield ""
    yield ""

    yield "@schema(title='Parameters')"
    yield "@dataclass"
    yield "class ModelLoadSaveDetectorArguments(Mapping[str, Any]):"
    yield "    filename: float = field("
    yield "    metadata=schema(title='filename', description='Filename to load/save.')"
    yield ")"
    yield ""

    yield "@dataclass"
    yield "class ModelLoadDetector:"
    yield "    name: str"
    yield "    arguments: ModelLoadSaveDetectorArguments"
    yield "    func: Literal['pyxel.models.load_detector'] = 'pyxel.models.load_detector'"
    yield "    enabled: bool = True"
    yield ""

    yield "@dataclass"
    yield "class ModelSaveDetector:"
    yield "    name: str"
    yield "    arguments: ModelLoadSaveDetectorArguments"
    yield "    func: Literal['pyxel.models.save_detector'] = 'pyxel.models.save_detector'"
    yield "    enabled: bool = True"
    yield ""

    yield "@dataclass"
    yield "class ModelFunction:"
    yield "    name: str"
    yield "    func: str = field(metadata=schema(pattern='^(?!pyxel\\.models\\.)'))"
    yield "    arguments: Optional[Mapping[str, Any]] = None"
    yield "    enabled: bool = True"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class ModelGroup:"
    yield "    models: Sequence[ModelFunction]"
    yield "    name: str"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class DetectionPipeline:"
    for model_name in DetectionPipeline.MODEL_GROUPS:
        yield f"    {model_name}: Optional[Sequence[ModelFunction]] = None"

    yield ""
    yield ""

    yield from generate_group(lst)
    all_configurations = yield from generate_detectors()

    yield "class NotEqualError(Exception):"
    yield "    ..."
    yield ""
    yield "def compare(first, second) -> None:"
    yield "    if type(first) != type(second):"
    yield "        raise NotEqualError"
    yield ""
    yield "    if isinstance(first, dict) and isinstance(second, dict):"
    yield "        if set(first) != set(second):"
    yield "            raise NotEqualError"
    yield ""
    yield "        for key in first:"
    yield "            compare(first[key], second[key])"
    yield ""
    yield "    elif isinstance(first, list) and isinstance(second, list):"
    yield "        if len(first) != len(second):"
    yield "            raise NotEqualError"
    yield ""
    yield "        if first != second:"
    yield "            sorted_first = sorted(first)"
    yield "            sorted_second = sorted(second)"
    yield ""
    yield "            if sorted_first != sorted_second:"
    yield "                raise NotEqualError"
    yield "    else:"
    yield "        if first != second:"
    yield "            raise NotEqualError"
    yield ""
    yield ""
    yield "@click.command()"
    yield "@click.option("
    yield "    '-f',"
    yield "    '--filename',"
    yield "    default='../../static/pyxel_schema.json',"
    yield "    type=click.Path(),"
    yield "    help='JSON schema filename',"
    yield "    show_default=True,"
    yield ")"

    yield "@click.option('--check', is_flag=True,"
    yield '     help="Don\'t write the JSON Schema back, just return the status.")'

    yield "def create_json_schema(filename: pathlib.Path, check: bool):"
    yield "    if sys.version_info < (3, 10):"
    yield "        raise NotImplementedError('This script must run on Python 3.10+')"
    yield ""
    yield "    # Manually define a 'format' for JSON Schema for 'Path'"
    yield "    schema(format='uri')(Path)"
    yield ""
    yield "    dct_schema = deserialization_schema("
    yield f"        Union[{','.join(all_configurations)}], version=JsonSchemaVersion.DRAFT_7, all_refs=True"
    yield "    )"
    yield ""
    yield ""
    yield "    full_filename = pathlib.Path(filename).resolve()"
    yield ""
    yield "    if check:"
    yield "        with full_filename.open() as fh:"
    yield "            dct_reference = json.load(fh)"
    yield ""
    yield "        new_dct_schema: Mapping[str, Any] = json.loads(json.dumps(dct_schema))"
    yield ""
    yield "        try:"
    yield "           compare(dct_reference, new_dct_schema)"
    yield "        except NotEqualError:"
    yield "            print("
    yield '                f"Error, JSON Schema file: {full_filename} is not the newest version. "'
    yield "                f\"Please run 'tox -e json_schema'\""
    yield "             )"
    yield "            pprint(result)"
    yield "            sys.exit(1)"
    yield "        else:"
    yield "            sys.exit(0)"
    yield "    else:"
    yield "        print(json.dumps(dct_schema))"
    yield "        with full_filename.open('w') as fh:"
    yield "            json.dump(obj=dct_schema, fp=fh, indent=2, sort_keys=True)"
    yield ""
    yield ""
    yield "if __name__ == '__main__':"
    yield "    create_json_schema()"


def create_auto_generated(filename: Path) -> None:
    """Create an auto-generated file."""
    with Path(filename).open("w") as fh:
        for line in tqdm(generate_all_models()):
            fh.write(f"{line}\n")


@click.command()
@click.option(
    "-f",
    "--filename",
    default="./auto_generated.py",
    type=click.Path(),
    help="Auto generated filename.",
    show_default=True,
)
@click.version_option(version=__version__)
def main(filename: Path):
    """Create an auto-generated file."""
    create_auto_generated(filename)


if __name__ == "__main__":
    main()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import numpy as np

from pyxel.util import set_random_seed


def test_set_random_seed_no_seed():
    """Test function 'set_random_seed' without a seed."""
    np.random.seed(123)
    exp_value1_seed123 = np.random.random()
    exp_value2_seed123 = np.random.random()
    exp_value3_seed123 = np.random.random()

    # Test before 'set_random_seed'
    np.random.seed(123)
    value = np.random.random()
    assert value == exp_value1_seed123

    # Test with 'set_random_seed'
    with set_random_seed(seed=None):
        value = np.random.random()
        assert value == exp_value2_seed123

    # Test after 'set_random_seed'
    value = np.random.random()
    assert value == exp_value3_seed123


def test_set_random_seed_with_seed456():
    """Test function 'set_random_seed' without a seed."""
    np.random.seed(123)
    exp_value1_seed123 = np.random.random()
    exp_value2_seed123 = np.random.random()

    # Test before 'set_random_seed'
    np.random.seed(456)
    exp_value1_seed456 = np.random.random()

    # Set first seed
    np.random.seed(123)
    value = np.random.random()
    assert value == exp_value1_seed123

    # Test with 'set_random_seed'
    with set_random_seed(seed=456):
        value = np.random.random()
        assert value == exp_value1_seed456

    # Test after 'set_random_seed'
    value = np.random.random()
    assert value == exp_value2_seed123


def test_set_random_seed_with_seed123():
    """Test function 'set_random_seed' without a seed."""
    np.random.seed(123)
    exp_value1_seed123 = np.random.random()
    exp_value2_seed123 = np.random.random()

    # Test before 'set_random_seed'
    np.random.seed(123)
    value = np.random.random()
    assert value == exp_value1_seed123

    # Test with 'set_random_seed'
    with set_random_seed(seed=123):
        value = np.random.random()
        assert value == exp_value1_seed123

    # Test after 'set_random_seed'
    value = np.random.random()
    assert value == exp_value2_seed123



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Parametric mode class and helper functions."""
import itertools
import logging
from collections import Counter
from collections.abc import Iterable, Iterator, Mapping, Sequence
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from functools import partial, reduce
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import dask.bag as db
import numpy as np
import pandas as pd
import toolz
from numpy.typing import ArrayLike
from tqdm.auto import tqdm

from pyxel.exposure import Readout, run_exposure_pipeline, run_pipeline
from pyxel.observation.parameter_values import ParameterType, ParameterValues
from pyxel.pipelines import ResultId, get_result_id
from pyxel.state import get_obj_att, get_value

if TYPE_CHECKING:
    import xarray as xr
    from datatree import DataTree

    from pyxel.outputs import ObservationOutputs
    from pyxel.pipelines import Processor


class ParameterMode(Enum):
    """Parameter mode class."""

    Product = "product"
    Sequential = "sequential"
    Custom = "custom"


class ObservationResult(NamedTuple):
    """Result class for observation class."""

    dataset: Union["xr.Dataset", dict[str, "xr.Dataset"]]
    parameters: "xr.Dataset"
    logs: "xr.Dataset"


@dataclass(frozen=True)
class ParameterItem:
    """Internal Parameter Item."""

    # TODO: Merge 'index' and 'parameters'
    index: tuple[int, ...]
    parameters: Mapping[str, Any]
    run_index: int


@dataclass(frozen=True)
class CustomParameterItem:
    """Internal Parameter Item."""

    # TODO: Merge 'index' and 'parameters'
    index: int
    parameters: Mapping[str, Any]
    run_index: int


def _get_short_name_with_model(name: str) -> str:
    _, _, model_name, _, param_name = name.split(".")

    return f"{model_name}.{param_name}"


def _get_final_short_name(name: str, param_type: ParameterType) -> str:
    if param_type == ParameterType.Simple:
        return name
    elif param_type == ParameterType.Multi:
        return _id(name)
    else:
        raise NotImplementedError


# TODO: This function will be deprecated (see #563)
def _get_short_dimension_names(types: Mapping[str, ParameterType]) -> Mapping[str, str]:
    # Create potential names for the dimensions
    potential_dim_names: dict[str, str] = {}
    for param_name, param_type in types.items():
        short_name: str = short(param_name)

        potential_dim_names[param_name] = _get_final_short_name(
            name=short_name, param_type=param_type
        )

    # Find possible duplicates
    count_dim_names: Mapping[str, int] = Counter(potential_dim_names.values())

    duplicate_dim_names: Sequence[str] = [
        name for name, freq in count_dim_names.items() if freq > 1
    ]

    if duplicate_dim_names:
        dim_names: dict[str, str] = {}
        for param_name, param_type in types.items():
            short_name = potential_dim_names[param_name]

            if short_name in duplicate_dim_names:
                new_short_name: str = _get_short_name_with_model(param_name)
                dim_names[param_name] = _get_final_short_name(
                    name=new_short_name, param_type=param_type
                )

            else:
                dim_names[param_name] = short_name

        return dim_names

    return potential_dim_names


# TODO: Add unit tests
def _get_short_dimension_names_new(
    types: Mapping[str, ParameterType]
) -> Mapping[str, str]:
    # Create potential names for the dimensions
    potential_dim_names: dict[str, str] = {}
    for param_name in types:
        short_name: str = short(param_name)

        potential_dim_names[param_name] = short_name

    # Find possible duplicates
    count_dim_names: Mapping[str, int] = Counter(potential_dim_names.values())

    duplicate_dim_names: Sequence[str] = [
        name for name, freq in count_dim_names.items() if freq > 1
    ]

    if duplicate_dim_names:
        dim_names: dict[str, str] = {}
        for param_name in types:
            short_name = potential_dim_names[param_name]

            if short_name in duplicate_dim_names:
                new_short_name: str = _get_short_name_with_model(param_name)
                dim_names[param_name] = new_short_name

            else:
                dim_names[param_name] = short_name

        return dim_names

    return potential_dim_names


class Observation:
    """Observation class."""

    def __init__(
        self,
        outputs: "ObservationOutputs",
        parameters: Sequence[ParameterValues],
        readout: Optional[Readout] = None,
        mode: Literal["product", "sequential", "custom"] = "product",
        from_file: Optional[str] = None,  # Note: Only For 'custom' mode
        column_range: Optional[tuple[int, int]] = None,  # Note: Only For 'custom' mode
        with_dask: bool = False,
        result_type: str = "all",
        pipeline_seed: Optional[int] = None,
    ):
        self.outputs = outputs
        self.readout: Readout = readout or Readout()
        self.parameter_mode: ParameterMode = ParameterMode(mode)
        self._parameters: Sequence[ParameterValues] = parameters

        # Specific to mode 'custom'
        self._custom_file: Optional[str] = from_file
        self._custom_data: Optional[pd.DataFrame] = None
        self._custom_columns: Optional[slice] = (
            slice(*column_range) if column_range else None
        )

        self.with_dask = with_dask
        self.parameter_types: dict[str, ParameterType] = {}
        self._result_type: ResultId = get_result_id(result_type)
        self._pipeline_seed = pipeline_seed

        if self.parameter_mode == ParameterMode.Custom:
            self._load_custom_parameters()

    def __repr__(self):
        cls_name: str = self.__class__.__name__
        return f"{cls_name}<mode={self.parameter_mode!s}>"

    @property
    def result_type(self) -> ResultId:
        """TBW."""
        return self._result_type

    @result_type.setter
    def result_type(self, value: ResultId) -> None:
        """TBW."""
        self._result_type = value

    @property
    def pipeline_seed(self) -> Optional[int]:
        """TBW."""
        return self._pipeline_seed

    @pipeline_seed.setter
    def pipeline_seed(self, value: int) -> None:
        """TBW."""
        self._pipeline_seed = value

    @property
    def enabled_steps(self) -> Sequence[ParameterValues]:
        """Return a list of enabled ParameterValues."""
        out = [step for step in self._parameters if step.enabled]
        return out

    # TODO: is self._custom_data really needed?
    def _load_custom_parameters(self) -> None:
        """Load custom parameters from file."""
        from pyxel import load_table

        if self._custom_file is None:
            raise ValueError("File for custom parametric mode not specified!")

        # Read the file without forcing its data type
        all_data: pd.DataFrame = load_table(self._custom_file, dtype=None)
        filtered_data: pd.DataFrame = all_data.loc[:, self._custom_columns]

        self._custom_data = filtered_data

    def _custom_parameters(
        self,
    ) -> Iterator[
        tuple[
            int,
            dict[str, Union[Number, str, Sequence[Union[Number, str]]]],
        ]
    ]:
        """Generate custom mode parameters based on input file.

        Yields
        ------
        index: int
        parameter_dict: dict
        """
        if not isinstance(self._custom_data, pd.DataFrame):
            raise TypeError("Custom parameters not loaded from file.")

        index: int
        row_serie: pd.Series
        for index, row_serie in self._custom_data.iterrows():
            row: Sequence[Union[Number, str]] = row_serie.to_list()

            i: int = 0
            parameter_dict: dict[
                str, Union[Number, str, Sequence[Union[Number, str]]]
            ] = {}
            for step in self.enabled_steps:
                key: str = step.key

                # TODO: this is confusing code. Fix this.
                #       Furthermore 'step.values' should be a `List[int, float]` and not a `str`
                if step.values == "_":
                    parameter_dict[key] = row[i]

                elif isinstance(step.values, Sequence):
                    values: np.ndarray = np.array(step.values)
                    values_flattened = values.flatten()

                    # TODO: find a way to remove the ignore
                    if not all(x == "_" for x in values_flattened):
                        raise ValueError(
                            'Only "_" characters (or a list of them) should be used to '
                            "indicate parameters updated from file in custom mode"
                        )

                    value: Sequence[Union[Number, str]] = row[
                        i : i + len(values_flattened)
                    ]
                    assert len(value) == len(step.values)

                    parameter_dict[key] = value

                else:
                    raise NotImplementedError

                i += len(step.values)

            yield index, parameter_dict

    def _sequential_parameters(self) -> Iterator[tuple[int, dict]]:
        """Generate sequential mode parameters.

        Yields
        ------
        index: int
        parameter_dict: dict
        """
        step: ParameterValues
        for step in self.enabled_steps:
            key: str = step.key
            for index, value in enumerate(step):
                parameter_dict = {key: value}
                yield index, parameter_dict

    def _product_indices(self) -> "Iterator[Tuple]":
        """Return an iterator of product parameter indices.

        Returns
        -------
        iterator
        """
        step_ranges = [range(len(step)) for step in self.enabled_steps]
        out = itertools.product(*step_ranges)
        return out

    def _product_parameters(
        self,
    ) -> Iterator[tuple[tuple, dict[str, Any]]]:
        """Generate product mode parameters.

        Yields
        ------
        indices: tuple
        parameter_dict: dict
        """
        all_steps = self.enabled_steps
        keys = [step.key for step in self.enabled_steps]
        for indices, params in zip(
            self._product_indices(), itertools.product(*all_steps)
        ):
            parameter_dict = {}
            for key, value in zip(keys, params):
                parameter_dict.update({key: value})
            yield indices, parameter_dict

    # TODO: This function will be deprecated (see #563)
    def _parameter_it(self) -> Iterator[tuple]:
        """Return the method for generating parameters based on parametric mode."""
        if self.parameter_mode == ParameterMode.Product:
            yield from self._product_parameters()

        elif self.parameter_mode == ParameterMode.Sequential:
            yield from self._sequential_parameters()

        elif self.parameter_mode == ParameterMode.Custom:
            yield from self._custom_parameters()
        else:
            raise NotImplementedError

    def _get_parameters_item(
        self, processor: "Processor"
    ) -> Sequence[Union[ParameterItem, CustomParameterItem]]:
        if self.parameter_mode == ParameterMode.Product:
            params_it: Iterator = self._product_parameters()

            return [
                ParameterItem(index=index, parameters=parameter_dict, run_index=n)
                for n, (index, parameter_dict) in enumerate(params_it)
            ]

        elif self.parameter_mode == ParameterMode.Sequential:
            # Get default values for all unique parameters
            params_all_keys: Sequence[str] = [
                param_value.key for param_value in self.enabled_steps
            ]
            params_unique_keys: Iterator[str] = toolz.unique(params_all_keys)

            params_defaults: Mapping[str, np.ndarray] = {
                key: processor.get(key) for key in params_unique_keys
            }

            params_it = self._sequential_parameters()

            return [
                CustomParameterItem(
                    index=index,
                    parameters=params_defaults | parameter_dict,
                    run_index=n,
                )
                for n, (index, parameter_dict) in enumerate(params_it)
            ]

        elif self.parameter_mode == ParameterMode.Custom:
            params_it = self._custom_parameters()

            return [
                CustomParameterItem(index=index, parameters=parameter_dict, run_index=n)
                for n, (index, parameter_dict) in enumerate(params_it)
            ]

        else:
            raise ValueError("Parametric mode not specified.")

    def _processors_it(
        self, processor: "Processor"
    ) -> Iterator[tuple["Processor", Union[int, tuple[int]], dict]]:
        """Generate processors with different product parameters.

        Parameters
        ----------
        processor: Processor

        Yields
        ------
        new_processor: Processor
        index: tuple of int
        parameter_dict: dict
        """

        for index, parameter_dict in self._parameter_it():
            new_processor = create_new_processor(
                processor=processor,
                parameter_dict=parameter_dict,
            )
            yield new_processor, index, parameter_dict

    def _get_parameter_types(self) -> Mapping[str, ParameterType]:
        """Check for each step if parameters can be used as dataset coordinates (1D, simple) or not (multi)."""
        for step in self.enabled_steps:
            self.parameter_types.update({step.key: step.type})
        return self.parameter_types

    def run_debug_mode(
        self, processor: "Processor"
    ) -> tuple[list["Processor"], "xr.Dataset"]:
        """Run observation pipelines in debug mode and return list of processors and parameter logs.

        Parameters
        ----------
        processor: Processor

        Returns
        -------
        processors: list
        final_logs: Dataset
        """
        # Late import to speedup start-up time
        import xarray as xr

        processors = []
        logs: list[xr.Dataset] = []

        for processor_id, (proc, _index, parameter_dict) in enumerate(
            tqdm(self._processors_it(processor))
        ):
            log: xr.Dataset = log_parameters(
                processor_id=processor_id, parameter_dict=parameter_dict
            )
            logs.append(log)
            _ = run_exposure_pipeline(
                processor=proc,
                readout=self.readout,
                outputs=self.outputs,
                pipeline_seed=self.pipeline_seed,
            )
            processors.append(processor)

        # See issue #276
        final_logs = xr.combine_by_coords(logs)
        if not isinstance(final_logs, xr.Dataset):
            raise TypeError("Expecting 'Dataset'.")

        return processors, final_logs

    def validate_steps(self, processor: "Processor") -> None:
        """Validate enabled parameter steps in processor before running the pipelines.

        Parameters
        ----------
        processor: Processor

        Raises
        ------
        KeyError
            If a 'step' is missing in the configuration.
        ValueError
            If a model referenced in the configuration has not been enabled.
        """
        step: ParameterValues
        for step in self.enabled_steps:
            key: str = step.key
            if not processor.has(key):
                raise KeyError(f"Missing parameter: {key!r} in steps.")

            # TODO: the string literal expressions are difficult to maintain.
            #     Example: 'pipeline.', '.arguments', '.enabled'
            #     We may want to consider an API for this.
            # Proposed API:
            # value = operator.attrgetter(step.key)(processor)
            if "pipeline." in key:
                model_name: str = key[: key.find(".arguments")]
                model_enabled: str = model_name + ".enabled"
                if not processor.get(model_enabled):
                    raise ValueError(
                        f"The '{model_name}' model referenced in Observation configuration "
                        f"has not been enabled in yaml config!"
                    )

            if (
                any(x == "_" for x in step.values[:])
                and self.parameter_mode != ParameterMode.Custom
            ):
                raise ValueError(
                    "Either define 'custom' as parametric mode or "
                    "do not use '_' character in 'values' field"
                )

    # TODO: This method will be deprecated (see #563)
    # ruff: noqa: C901
    def run_observation(self, processor: "Processor") -> ObservationResult:
        """Run the observation pipelines.

        Parameters
        ----------
        processor : Processor

        Returns
        -------
        Result
        """
        # Late import to speedup start-up time
        import xarray as xr

        # validation
        self.validate_steps(processor)

        types: Mapping[str, ParameterType] = self._get_parameter_types()

        dim_names: Mapping[str, str] = _get_short_dimension_names(types)

        y = range(processor.detector.geometry.row)
        x = range(processor.detector.geometry.col)
        times = self.readout.times

        if self.parameter_mode == ParameterMode.Product:
            apply_pipeline = partial(
                self._apply_exposure_pipeline_product,
                dimension_names=dim_names,
                x=x,
                y=y,
                processor=processor,
                times=times,
                types=types,
            )
            lst = [
                (index, parameter_dict, n)
                for n, (index, parameter_dict) in enumerate(self._parameter_it())
            ]

            if self.with_dask:
                dataset_list = db.from_sequence(lst).map(apply_pipeline).compute()
            else:
                dataset_list = list(map(apply_pipeline, tqdm(lst)))

            # prepare lists for to-be-merged datasets
            parameters: list[list[xr.Dataset]] = [
                [] for _ in range(len(self.enabled_steps))
            ]
            logs = []

            for processor_id, (indices, parameter_dict, _) in enumerate(lst):
                # log parameters for this pipeline
                log = log_parameters(
                    processor_id=processor_id, parameter_dict=parameter_dict
                )
                logs.append(log)

                # save parameters with appropriate product mode indices
                for i, coordinate in enumerate(parameter_dict):
                    parameter_ds = parameter_to_dataset(
                        parameter_dict=parameter_dict,
                        dimension_names=dim_names,
                        index=indices[i],
                        coordinate_name=coordinate,
                    )
                    parameters[i].append(parameter_ds)

            # merging/combining the outputs
            final_parameters_list: list[xr.Dataset] = []
            for p in parameters:
                # See issue #276
                new_dataset = xr.combine_by_coords(p)
                if not isinstance(new_dataset, xr.Dataset):
                    raise TypeError("Expecting 'Dataset'.")

                final_parameters_list.append(new_dataset)

            final_parameters_merged = xr.merge(final_parameters_list)

            # See issue #276
            final_logs = xr.combine_by_coords(logs)
            if not isinstance(final_logs, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")

            # See issue #276
            final_dataset = xr.combine_by_coords(dataset_list)
            if not isinstance(final_dataset, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")

            result = ObservationResult(
                dataset=final_dataset,
                parameters=final_parameters_merged,
                logs=final_logs,
            )

            return result

        elif self.parameter_mode == ParameterMode.Sequential:
            apply_pipeline = partial(
                self._apply_exposure_pipeline_sequential,
                dimension_names=dim_names,
                x=x,
                y=y,
                processor=processor,
                times=times,
                types=types,
            )

            lst = [
                (index, parameter_dict, n)
                for n, (index, parameter_dict) in enumerate(self._parameter_it())
            ]

            if self.with_dask:
                dataset_list = db.from_sequence(lst).map(apply_pipeline).compute()
            else:
                dataset_list = list(map(apply_pipeline, tqdm(lst)))

            # prepare lists/dictionaries for to-be-merged datasets
            parameters = [[] for _ in range(len(self.enabled_steps))]
            logs = []

            # overflow to next parameter step counter
            step_counter = -1

            for processor_id, (index, parameter_dict, _) in enumerate(lst):
                # log parameters for this pipeline
                # TODO: somehow refactor logger so that default parameters
                #  from other steps are also logged in sequential mode
                log = log_parameters(
                    processor_id=processor_id, parameter_dict=parameter_dict
                )
                logs.append(log)

                # Figure out current coordinate
                coordinate = str(next(iter(parameter_dict)))
                # Check for overflow to next parameter
                if index == 0:
                    step_counter += 1

                # save sequential parameter with appropriate index
                parameter_ds = parameter_to_dataset(
                    parameter_dict=parameter_dict,
                    dimension_names=dim_names,
                    index=index,
                    coordinate_name=coordinate,
                )
                parameters[step_counter].append(parameter_ds)

            # merging/combining the outputs
            # See issue #276
            final_logs = xr.combine_by_coords(logs)
            if not isinstance(final_logs, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")

            final_datasets = compute_final_sequential_dataset(
                list_of_index_and_parameter=lst,
                list_of_datasets=dataset_list,
                dimension_names=dim_names,
            )

            final_parameters_list = []
            for p in parameters:
                # See issue #276
                new_dataset = xr.combine_by_coords(p)
                if not isinstance(new_dataset, xr.Dataset):
                    raise TypeError("Expecting 'Dataset'.")

                final_parameters_list.append(new_dataset)

            final_parameters_merged = xr.merge(final_parameters_list)

            result = ObservationResult(
                dataset=final_datasets,
                parameters=final_parameters_merged,
                logs=final_logs,
            )
            return result

        elif self.parameter_mode == ParameterMode.Custom:
            apply_pipeline = partial(
                self._apply_exposure_pipeline_custom,
                x=x,
                y=y,
                processor=processor,
                times=times,
            )
            lst = [
                (index, parameter_dict, n)
                for n, (index, parameter_dict) in enumerate(self._parameter_it())
            ]

            if self.with_dask:
                dataset_list = db.from_sequence(lst).map(apply_pipeline).compute()
            else:
                dataset_list = list(map(apply_pipeline, tqdm(lst)))

            # prepare lists for to-be-merged datasets
            logs = []

            for index, parameter_dict, _ in lst:
                # log parameters for this pipeline
                log = log_parameters(processor_id=index, parameter_dict=parameter_dict)
                logs.append(log)

            # merging/combining the outputs
            final_ds = xr.combine_by_coords(dataset_list)
            final_log = xr.combine_by_coords(logs)

            # See issue #276
            if not isinstance(final_ds, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")
            if not isinstance(final_log, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")

            final_parameters = final_log  # parameter dataset same as logs

            result = ObservationResult(
                dataset=final_ds, parameters=final_parameters, logs=final_log
            )
            return result

        else:
            raise ValueError("Parametric mode not specified.")

    def run_observation_datatree(self, processor: "Processor") -> "DataTree":
        """Run the observation pipelines."""
        # Late import to speedup start-up time
        from datatree import DataTree

        # validation
        self.validate_steps(processor)

        types: Mapping[str, ParameterType] = self._get_parameter_types()
        dim_names: Mapping[str, str] = _get_short_dimension_names_new(types)

        apply_pipeline: Callable[
            [Union[ParameterItem, CustomParameterItem]], DataTree
        ] = partial(
            self._apply_exposure_pipeline_new,
            dimension_names=dim_names,
            processor=processor,
            types=types,
        )

        parameters: Sequence[
            Union[ParameterItem, CustomParameterItem]
        ] = self._get_parameters_item(processor=processor)

        if self.with_dask:
            datatree_bag: db.Bag = (
                db.from_sequence(parameters)
                .map(apply_pipeline)
                .fold(binop=DataTree.combine_first, combine=DataTree.combine_first)  # type: ignore
            )
            final_datatree: DataTree = datatree_bag.compute()
        else:
            datatree_list: Iterator[DataTree] = (
                apply_pipeline(el) for el in tqdm(parameters)
            )
            final_datatree = reduce(DataTree.combine_first, datatree_list)  # type: ignore

        parameter_name: str = self.parameter_mode.name
        final_datatree.attrs["running mode"] = f"Observation - {parameter_name}"

        # See issue #276. TODO: Is this still valid ?
        # if not isinstance(final_dataset, xr.Dataset):
        #     raise TypeError("Expecting 'Dataset'.")

        return final_datatree

    # TODO: This function will be deprecated (see #563)
    def _apply_exposure_pipeline_product(
        self,
        index_and_parameter: tuple[
            tuple[int, ...],
            Mapping[
                str,
                Union[str, Number, np.ndarray, list[Union[str, Number, np.ndarray]]],
            ],
            int,
        ],
        dimension_names: Mapping[str, str],
        processor: "Processor",
        x: range,
        y: range,
        times: np.ndarray,
        types: Mapping[str, ParameterType],
    ) -> "xr.Dataset":
        index, parameter_dict, n = index_and_parameter

        new_processor = create_new_processor(
            processor=processor, parameter_dict=parameter_dict
        )

        # run the pipeline
        _ = run_exposure_pipeline(
            processor=new_processor,
            readout=self.readout,
            result_type=self.result_type,
            pipeline_seed=self.pipeline_seed,
        )

        _ = self.outputs.save_to_file(processor=new_processor, run_number=n)

        ds: xr.Dataset = new_processor.result_to_dataset(
            x=x,
            y=y,
            times=times,
            result_type=self.result_type,
        )

        # Can also be done outside dask in a loop
        ds = _add_product_parameters(
            ds=ds,
            parameter_dict=parameter_dict,
            dimension_names=dimension_names,
            indices=index,
            types=types,
        )

        ds.attrs.update({"running mode": "Observation - Product"})

        return ds

    def _apply_exposure_pipeline_new(
        self,
        param_item: Union[ParameterItem, CustomParameterItem],
        dimension_names: Mapping[str, str],
        processor: "Processor",
        types: Mapping[str, ParameterType],
    ) -> "DataTree":
        new_processor = create_new_processor(
            processor=processor,
            parameter_dict=param_item.parameters,
        )

        # run the pipeline
        data_tree: "DataTree" = run_pipeline(
            processor=new_processor,
            readout=self.readout,
            result_type=self.result_type,
            pipeline_seed=self.pipeline_seed,
        )

        _ = self.outputs.save_to_file(
            processor=new_processor,
            run_number=param_item.run_index,
        )

        # Can also be done outside dask in a loop
        if isinstance(param_item, ParameterItem):
            final_data_tree = _add_product_parameters_datatree(
                data_tree=data_tree,
                parameter_dict=param_item.parameters,
                indexes=param_item.index,
                dimension_names=dimension_names,
                types=types,
            )

        else:
            final_data_tree = _add_custom_parameters_datatree(
                data_tree=data_tree,
                parameter_dict=param_item.parameters,
                index=param_item.index,
                dimension_names=dimension_names,
                types=types,
            )

        return final_data_tree

    # TODO: This function will be deprecated (see #563)
    def _apply_exposure_pipeline_custom(
        self,
        index_and_parameter: tuple[
            int,
            Mapping[
                str,
                Union[str, Number, np.ndarray, list[Union[str, Number, np.ndarray]]],
            ],
            int,
        ],
        processor: "Processor",
        x: range,
        y: range,
        times: np.ndarray,
    ):
        index, parameter_dict, n = index_and_parameter

        new_processor = create_new_processor(
            processor=processor,
            parameter_dict=parameter_dict,
        )

        # run the pipeline
        _ = run_exposure_pipeline(
            processor=new_processor,
            readout=self.readout,
            result_type=self.result_type,
            pipeline_seed=self.pipeline_seed,
        )

        _ = self.outputs.save_to_file(processor=new_processor, run_number=n)

        ds: xr.Dataset = new_processor.result_to_dataset(
            x=x, y=y, times=times, result_type=self.result_type
        )

        # Can also be done outside dask in a loop
        ds = _add_custom_parameters(
            ds=ds,
            index=index,
        )
        ds.attrs.update({"running mode": "Observation - Custom"})

        return ds

    # TODO: This function will be deprecated (see #563)
    def _apply_exposure_pipeline_sequential(
        self,
        index_and_parameter: tuple[
            int,
            Mapping[
                str,
                Union[str, Number, np.ndarray, list[Union[str, Number, np.ndarray]]],
            ],
            int,
        ],
        dimension_names: Mapping[str, str],
        processor: "Processor",
        x: range,
        y: range,
        times: np.ndarray,
        types: Mapping[str, ParameterType],
    ):
        index, parameter_dict, n = index_and_parameter

        new_processor = create_new_processor(
            processor=processor,
            parameter_dict=parameter_dict,
        )

        coordinate = str(next(iter(parameter_dict)))

        # run the pipeline
        _ = run_exposure_pipeline(
            processor=new_processor,
            readout=self.readout,
            result_type=self.result_type,
            pipeline_seed=self.pipeline_seed,
        )

        _ = self.outputs.save_to_file(processor=new_processor, run_number=n)

        ds: xr.Dataset = new_processor.result_to_dataset(
            x=x, y=y, times=times, result_type=self.result_type
        )

        # Can also be done outside dask in a loop
        ds = _add_sequential_parameters(
            ds=ds,
            parameter_dict=parameter_dict,
            dimension_names=dimension_names,
            index=index,
            coordinate_name=coordinate,
            types=types,
        )

        ds.attrs.update({"running mode": "Observation - Sequential"})

        return ds

    def debug_parameters(self, processor: "Processor") -> list:
        """List the parameters using processor parameters in processor generator.

        Parameters
        ----------
        processor: Processor

        Returns
        -------
        result: list
        """
        result = []
        processor_generator = self._processors_it(processor=processor)
        for i, (proc, _, _) in enumerate(processor_generator):
            values = []
            for step in self.enabled_steps:
                _, att = get_obj_att(proc, step.key)
                value = get_value(proc, step.key)
                values.append((att, value))
            logging.debug("%d: %r", i, values)
            result.append((i, values))
        return result


def create_new_processor(
    processor: "Processor",
    parameter_dict: Mapping[
        str, Union[str, Number, np.ndarray, list[Union[str, Number, np.ndarray]]]
    ],
) -> "Processor":
    """Create a copy of processor and set new attributes from a dictionary before returning it.

    Parameters
    ----------
    processor: Processor
    parameter_dict: dict

    Returns
    -------
    Processor
    """

    new_processor = deepcopy(processor)

    for key in parameter_dict.keys():
        new_processor.set(key=key, value=parameter_dict[key])

    return new_processor


def _id(s: str) -> str:
    """Add _id to the end of a string."""
    out = s + "_id"
    return out


def short(s: str) -> str:
    """Split string with . and return the last element."""
    out = s.split(".")[-1]
    return out


def log_parameters(processor_id: int, parameter_dict: dict) -> "xr.Dataset":
    """Return parameters in the current processor in a xarray dataset.

    Parameters
    ----------
    processor_id: int
    parameter_dict: dict

    Returns
    -------
    Dataset
    """
    # Late import to speedup start-up time
    import xarray as xr

    out = xr.Dataset()
    for key, value in parameter_dict.items():
        da = xr.DataArray(value)
        da = da.assign_coords(coords={"id": processor_id})
        da = da.expand_dims(dim="id")
        out[short(key)] = da
    return out


def parameter_to_dataset(
    parameter_dict: dict,
    dimension_names: Mapping[str, str],
    index: int,
    coordinate_name: str,
) -> "xr.Dataset":
    """Return a specific parameter dataset from a parameter dictionary.

    Parameters
    ----------
    parameter_dict: dict
    dimension_names
    index: int
    coordinate_name: str

    Returns
    -------
    Dataset
    """
    # Late import to speedup start-up time
    import xarray as xr

    parameter_ds = xr.Dataset()
    parameter = xr.DataArray(parameter_dict[coordinate_name])

    # TODO: Dirty hack. Fix this !
    short_name: str = dimension_names[coordinate_name]

    if short_name.endswith("_id"):
        short_coord_name = short_name[:-3]
        short_coord_name_id = short_name
    else:
        short_coord_name = short_name
        short_coord_name_id = f"{short_name}_id"

    parameter = parameter.assign_coords({short_coord_name_id: index})
    parameter = parameter.expand_dims(dim=short_coord_name_id)
    parameter_ds[short_coord_name] = parameter

    return parameter_ds


# TODO: This function will be deprecated (see #563)
def _add_custom_parameters(ds: "xr.Dataset", index: int) -> "xr.Dataset":
    """Add coordinate "index" to the dataset.

    Parameters
    ----------
    ds: Dataset
    index: int

    Returns
    -------
    Dataset
    """

    ds = ds.assign_coords({"id": index})
    ds = ds.expand_dims(dim="id")

    return ds


def _add_custom_parameters_datatree(
    data_tree: "DataTree",
    parameter_dict: Mapping[str, Union[str, Number, ArrayLike]],
    index: int,
    dimension_names: Mapping[str, str],
    types: Mapping[str, ParameterType],
) -> "DataTree":
    """Add coordinate "index" to the dataset.

    Parameters
    ----------
    data_tree: Dataset
    index: int

    Returns
    -------
    DataTree
    """
    import pandas as pd
    import xarray as xr

    data_tree = data_tree.expand_dims({"id": [index]})

    for coordinate_name, param_value in parameter_dict.items():
        short_name: str = dimension_names[coordinate_name]

        #  assigning the right coordinates based on type
        if types[coordinate_name] == ParameterType.Simple:
            data_tree = data_tree.assign_coords(
                {short_name: ("id", pd.Index([param_value]))}
            )

        elif types[coordinate_name] == ParameterType.Multi:
            data = np.array(param_value)
            data_array = xr.DataArray(data).expand_dims({"id": [index]})
            data_tree = data_tree.assign_coords({short_name: data_array})

        else:
            raise NotImplementedError

    return data_tree


# TODO: This function will be deprecated (see #563)
def _add_sequential_parameters(
    ds: "xr.Dataset",
    parameter_dict: Mapping[
        str, Union[str, Number, np.ndarray, list[Union[str, Number, np.ndarray]]]
    ],
    dimension_names: Mapping[str, str],
    index: int,
    coordinate_name: str,
    types: Mapping[str, ParameterType],
) -> "xr.Dataset":
    """Add true coordinates or index to sequential mode dataset.

    Parameters
    ----------
    ds : Dataset
    parameter_dict : dict
    dimension_names
    index : int
    coordinate_name : str
    types : dict

    Returns
    -------
    Dataset
    """

    #  assigning the right coordinates based on type
    short_name: str = dimension_names[coordinate_name]

    if types[coordinate_name] == ParameterType.Simple:
        ds = ds.assign_coords(coords={short_name: parameter_dict[coordinate_name]})
        ds = ds.expand_dims(dim=short_name)

    elif types[coordinate_name] == ParameterType.Multi:
        ds = ds.assign_coords({short_name: index})
        ds = ds.expand_dims(dim=short_name)

    return ds


# TODO: This function will be deprecated (see #563)
def _add_product_parameters(
    ds: "xr.Dataset",
    parameter_dict: Mapping[
        str, Union[str, Number, np.ndarray, list[Union[str, Number, np.ndarray]]]
    ],
    dimension_names: Mapping[str, str],
    indices: tuple[int, ...],
    types: Mapping[str, ParameterType],
) -> "xr.Dataset":
    """Add true coordinates or index to product mode dataset.

    Parameters
    ----------
    ds: Dataset
    parameter_dict: dict
    indices: tuple
    types: dict

    Returns
    -------
    Dataset
    """
    # TODO: Implement for coordinate 'multi'
    for i, (coordinate_name, param_value) in enumerate(parameter_dict.items()):
        short_name: str = dimension_names[coordinate_name]

        #  assigning the right coordinates based on type
        if types[coordinate_name] == ParameterType.Simple:
            ds = ds.assign_coords(coords={short_name: param_value})
            ds = ds.expand_dims(dim=short_name)

        elif types[coordinate_name] == ParameterType.Multi:
            ds = ds.assign_coords({short_name: indices[i]})
            ds = ds.expand_dims(dim=short_name)

        else:
            raise NotImplementedError

    return ds


def to_tuples(data: Iterable) -> tuple:
    lst: list = []
    for el in data:
        if isinstance(el, Iterable) and not isinstance(el, str):
            lst.append(to_tuples(el))
        else:
            lst.append(el)

    return tuple(lst)


def _add_product_parameters_datatree(
    data_tree: "DataTree",
    parameter_dict: Mapping[str, Union[str, Number, ArrayLike]],
    indexes: tuple[int, ...],
    dimension_names: Mapping[str, str],
    types: Mapping[str, ParameterType],
) -> "DataTree":
    """Add true coordinates or index to product mode dataset.

    Parameters
    ----------
    data_tree : DataTree
    parameter_dict : dict
    types : dict

    Returns
    -------
    DataTree
    """
    import xarray as xr

    # TODO: join 'indexes' and 'parameter_dict'
    for index, (coordinate_name, param_value) in zip(indexes, parameter_dict.items()):
        short_name: str = dimension_names[coordinate_name]

        #  assigning the right coordinates based on type
        if types[coordinate_name] == ParameterType.Simple:
            data_tree = data_tree.expand_dims(dim={short_name: [param_value]})

        elif types[coordinate_name] == ParameterType.Multi:
            data = np.array(param_value)
            data_array = xr.DataArray(data).expand_dims(
                dim={f"{short_name}_id": [index]}
            )
            data_tree = data_tree.expand_dims(
                {f"{short_name}_id": [index]}
            ).assign_coords({short_name: data_array})

        else:
            raise NotImplementedError

    return data_tree


def compute_final_sequential_dataset(
    list_of_index_and_parameter: list,
    list_of_datasets: list,
    dimension_names: Mapping[str, str],
) -> dict[str, "xr.Dataset"]:
    """Return a dictionary of result datasets where keys are different parameters.

    Parameters
    ----------
    list_of_index_and_parameter: list
    list_of_datasets: list
    dimension_names

    Returns
    -------
    dict
    """
    # Late import to speedup start-up time
    import xarray as xr

    final_dict: dict[str, list[xr.Dataset]] = {}

    for _, parameter_dict, n in list_of_index_and_parameter:
        coordinate = str(next(iter(parameter_dict)))
        coordinate_short: str = dimension_names[coordinate]

        if short(coordinate) not in final_dict:
            final_dict.update({coordinate_short: []})
            final_dict[coordinate_short].append(list_of_datasets[n])
        else:
            final_dict[coordinate_short].append(list_of_datasets[n])

    final_datasets: dict[str, xr.Dataset] = {}
    for key, value in final_dict.items():
        ds = xr.combine_by_coords(value)
        # see issue #276
        if not isinstance(ds, xr.Dataset):
            raise TypeError("Expecting 'Dataset'.")

        final_datasets.update({key: ds})

    return final_datasets



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from copy import deepcopy

import numpy as np
import pandas as pd
import pytest

from pyxel.detectors import MKID, Characteristics, Detector, Environment, MKIDGeometry


@pytest.fixture
def valid_mkid() -> MKID:
    """Create a valid `MKID`."""
    return MKID(
        geometry=MKIDGeometry(
            row=100,
            col=120,
            total_thickness=123.1,
            pixel_horz_size=12.4,
            pixel_vert_size=34.5,
        ),
        environment=Environment(temperature=100.1),
        characteristics=Characteristics(
            quantum_efficiency=0.1,
            charge_to_volt_conversion=0.2,
            pre_amplification=3.3,
        ),
    )


@pytest.mark.parametrize(
    "other_obj, is_equal",
    [
        pytest.param(None, False, id="None"),
        pytest.param(
            MKID(
                geometry=MKIDGeometry(row=100, col=120),
                environment=Environment(),
                characteristics=Characteristics(),
            ),
            False,
            id="Empty 'MKID'",
        ),
        pytest.param(
            MKID(
                geometry=MKIDGeometry(
                    row=100,
                    col=120,
                    total_thickness=123.1,
                    pixel_horz_size=12.4,
                    pixel_vert_size=34.5,
                ),
                environment=Environment(),
                characteristics=Characteristics(),
            ),
            False,
            id="Almost same parameters, different class",
        ),
        pytest.param(
            MKID(
                geometry=MKIDGeometry(
                    row=100,
                    col=120,
                    total_thickness=123.1,
                    pixel_horz_size=12.4,
                    pixel_vert_size=34.5,
                ),
                environment=Environment(temperature=100.1),
                characteristics=Characteristics(
                    quantum_efficiency=0.1,
                    charge_to_volt_conversion=0.2,
                    pre_amplification=3.3,
                ),
            ),
            True,
            id="Same parameters, same class",
        ),
    ],
)
def test_is_equal(valid_mkid: MKID, other_obj, is_equal):
    """Test equality statement for `MKID`."""

    if is_equal:
        assert valid_mkid == other_obj
    else:
        assert valid_mkid != other_obj


def test_is_equal_with_arrays(valid_mkid: MKID):
    other_detector = deepcopy(valid_mkid)

    assert valid_mkid.geometry is not other_detector.geometry
    assert valid_mkid.environment is not other_detector.environment
    assert valid_mkid.characteristics is not other_detector.characteristics
    assert valid_mkid._photon is not other_detector._photon
    assert valid_mkid._charge is not other_detector._charge
    assert valid_mkid._pixel is not other_detector._pixel
    assert valid_mkid._signal is not other_detector._signal
    assert valid_mkid._image is not other_detector._image
    assert valid_mkid._phase is not other_detector._phase

    shape = valid_mkid.geometry.row, valid_mkid.geometry.col
    photon: np.ndarray = np.random.random(size=shape)
    pixel: np.ndarray = np.random.random(size=shape)
    signal: np.ndarray = np.random.random(size=shape)
    image: np.ndarray = np.random.random(size=shape)
    charge: np.ndarray = np.random.random(size=shape)
    phase: np.ndarray = np.random.random(size=shape)

    valid_mkid.photon.array = photon.copy()
    valid_mkid.pixel.array = pixel.copy()
    valid_mkid.signal.array = signal.copy()
    valid_mkid.image.array = image.copy()
    valid_mkid.phase.array = phase.copy()

    valid_mkid.charge._array = charge.copy()
    valid_mkid.charge._frame = pd.concat(
        [valid_mkid.charge._frame, pd.DataFrame({"charge": [1]})],
        ignore_index=True,
    )

    other_detector.photon.array = photon.copy()
    other_detector.pixel.array = pixel.copy()
    other_detector.signal.array = signal.copy()
    other_detector.image.array = image.copy()
    other_detector.phase.array = phase.copy()

    other_detector.charge._array = charge.copy()
    other_detector.charge._frame = pd.concat(
        [other_detector.charge._frame, pd.DataFrame({"charge": [1]})],
        ignore_index=True,
    )

    assert valid_mkid == other_detector


def comparison(dct, other_dct):
    assert set(dct) == set(other_dct) == {"version", "type", "data", "properties"}
    assert dct["version"] == other_dct["version"]
    assert dct["type"] == other_dct["type"]
    assert dct["properties"] == other_dct["properties"]

    assert (
        set(dct["data"])
        == set(other_dct["data"])
        == {
            "photon",
            "scene",
            "pixel",
            "signal",
            "image",
            "phase",
            "charge",
            "data",
        }
    )
    np.testing.assert_equal(dct["data"]["photon"], other_dct["data"]["photon"])
    np.testing.assert_equal(dct["data"]["pixel"], other_dct["data"]["pixel"])
    np.testing.assert_equal(dct["data"]["signal"], other_dct["data"]["signal"])
    np.testing.assert_equal(dct["data"]["image"], other_dct["data"]["image"])
    np.testing.assert_equal(dct["data"]["phase"], other_dct["data"]["phase"])

    assert (
        set(dct["data"]["charge"])
        == set(other_dct["data"]["charge"])
        == {"array", "frame"}
    )
    np.testing.assert_equal(
        dct["data"]["charge"]["array"], other_dct["data"]["charge"]["array"]
    )
    pd.testing.assert_frame_equal(
        dct["data"]["charge"]["frame"], other_dct["data"]["charge"]["frame"]
    )


@pytest.mark.parametrize("klass", [MKID, Detector])
@pytest.mark.parametrize(
    "obj, exp_dict",
    [
        pytest.param(
            MKID(
                geometry=MKIDGeometry(row=100, col=120),
                environment=Environment(),
                characteristics=Characteristics(),
            ),
            {
                "version": 1,
                "type": "MKID",
                "properties": {
                    "geometry": {
                        "row": 100,
                        "col": 120,
                        "total_thickness": None,
                        "pixel_horz_size": None,
                        "pixel_vert_size": None,
                    },
                    "environment": {"temperature": None},
                    "characteristics": {
                        "quantum_efficiency": None,
                        "charge_to_volt_conversion": None,
                        "pre_amplification": None,
                        "full_well_capacity": None,
                        "adc_bit_resolution": None,
                        "adc_voltage_range": None,
                    },
                },
                "data": {
                    "photon": np.zeros(shape=(100, 120)),
                    "scene": None,
                    "pixel": np.zeros(shape=(100, 120)),
                    "signal": np.zeros(shape=(100, 120)),
                    "image": np.zeros(shape=(100, 120)),
                    "phase": np.zeros(shape=(100, 120)),
                    "charge": {
                        "array": np.zeros(shape=(100, 120)),
                        "frame": pd.DataFrame(
                            columns=[
                                "charge",
                                "number",
                                "init_energy",
                                "energy",
                                "init_pos_ver",
                                "init_pos_hor",
                                "init_pos_z",
                                "position_ver",
                                "position_hor",
                                "position_z",
                                "velocity_ver",
                                "velocity_hor",
                                "velocity_z",
                            ],
                            dtype=float,
                        ),
                    },
                    "data": {},
                },
            },
            id="Default Parameters",
        ),
        pytest.param(
            MKID(
                geometry=MKIDGeometry(
                    row=100,
                    col=120,
                    total_thickness=123.1,
                    pixel_horz_size=12.4,
                    pixel_vert_size=34.5,
                ),
                environment=Environment(temperature=100.1),
                characteristics=Characteristics(
                    quantum_efficiency=0.1,
                    charge_to_volt_conversion=0.2,
                    pre_amplification=3.3,
                    adc_bit_resolution=16,
                    adc_voltage_range=(0.0, 10.0),
                ),
            ),
            {
                "version": 1,
                "type": "MKID",
                "properties": {
                    "geometry": {
                        "row": 100,
                        "col": 120,
                        "total_thickness": 123.1,
                        "pixel_horz_size": 12.4,
                        "pixel_vert_size": 34.5,
                    },
                    "environment": {"temperature": 100.1},
                    "characteristics": {
                        "quantum_efficiency": 0.1,
                        "charge_to_volt_conversion": 0.2,
                        "pre_amplification": 3.3,
                        "full_well_capacity": None,
                        "adc_bit_resolution": 16,
                        "adc_voltage_range": (0.0, 10.0),
                    },
                },
                "data": {
                    "photon": np.zeros(shape=(100, 120)),
                    "scene": None,
                    "pixel": np.zeros(shape=(100, 120)),
                    "signal": np.zeros(shape=(100, 120)),
                    "image": np.zeros(shape=(100, 120)),
                    "phase": np.zeros(shape=(100, 120)),
                    "charge": {
                        "array": np.zeros(shape=(100, 120)),
                        "frame": pd.DataFrame(
                            columns=[
                                "charge",
                                "number",
                                "init_energy",
                                "energy",
                                "init_pos_ver",
                                "init_pos_hor",
                                "init_pos_z",
                                "position_ver",
                                "position_hor",
                                "position_z",
                                "velocity_ver",
                                "velocity_hor",
                                "velocity_z",
                            ],
                            dtype=float,
                        ),
                    },
                    "data": {},
                },
            },
            id="MKID fully defined",
        ),
    ],
)
def test_to_and_from_dict(klass, obj, exp_dict):
    """Test methods 'to_dict', 'from_dict'."""
    assert type(obj) == MKID

    # Convert from `MKID` to a `dict`
    dct = obj.to_dict()
    comparison(dct, exp_dict)

    # Copy 'exp_dict'
    copied_dct = deepcopy(exp_dict)  # create a new dict
    assert copied_dct is not exp_dict
    comparison(copied_dct, exp_dict)

    # Convert from `dict` to `MKID`
    other_obj = klass.from_dict(copied_dct)
    assert type(other_obj) == MKID
    assert obj == other_obj
    assert obj is not other_obj
    comparison(copied_dct, exp_dict)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import logging
import operator
from collections.abc import Sequence
from copy import deepcopy
from numbers import Number
from typing import TYPE_CHECKING, Any, Callable, NewType, Optional, Union

import numpy as np

from pyxel import __version__
from pyxel.evaluator import eval_entry
from pyxel.pipelines import DetectionPipeline, ModelGroup
from pyxel.state import get_obj_att
from pyxel.util.memory import get_size

if TYPE_CHECKING:
    import xarray as xr

    from pyxel.detectors import Detector


# class ResultType(Enum):
#     """Result type class."""
#
#     Photon = "photon"
#     Charge = "charge"
#     Pixel = "pixel"
#     Signal = "signal"
#     Image = "image"
#     Data = "data"
#     All = "all"

ResultId = NewType("ResultId", str)


def get_result_id(name: str) -> ResultId:
    """Convert to a 'ResultId' object."""
    if name not in (
        "scene",
        "photon",
        "charge",
        "pixel",
        "signal",
        "image",
        "data",
        "all",
    ) and not name.startswith("data"):
        raise ValueError(f"Result type: {name!r} unknown !")

    return ResultId(name)


def result_keys(result_type: ResultId) -> Sequence[ResultId]:
    """Return result keys based on result type.

    Parameters
    ----------
    result_type

    Returns
    -------
    list
    """
    if result_type == "all":
        return [
            ResultId("scene"),
            ResultId("photon"),
            ResultId("charge"),
            ResultId("pixel"),
            ResultId("signal"),
            ResultId("image"),
            ResultId("data"),
        ]

    return [ResultId(result_type)]


# TODO: Is this class needed ?
class Processor:
    """TBW."""

    def __init__(self, detector: "Detector", pipeline: DetectionPipeline):
        self._log = logging.getLogger(__name__)

        self.detector = detector
        self.pipeline = pipeline
        self._result: Optional[dict] = None  # TODO: Deprecate this variable ?

        self._numbytes = 0

    def __repr__(self) -> str:
        cls_name: str = self.__class__.__name__
        return f"{cls_name}<detector={self.detector!r}, pipeline={self.pipeline!r}>"

    def __deepcopy__(self, memodict: dict) -> "Processor":
        return Processor(
            detector=deepcopy(self.detector, memo=memodict),
            pipeline=deepcopy(self.pipeline, memo=memodict),
        )

    # TODO: Could it be renamed '__contains__' ?
    # TODO: reimplement this method.
    def has(self, key: str) -> bool:
        """Check if a parameter is available in this Processor.

        Examples
        --------
        >>> processor = Processor(...)
        >>> processor.has("pipeline.photon_collection.illumination.arguments.level")
        True
        """
        found = False
        obj, att = get_obj_att(self, key)
        if isinstance(obj, dict) and att in obj:
            found = True
        elif hasattr(obj, att):
            found = True
        return found

    # TODO: Could it be renamed '__getitem__' ?
    # TODO: Is it really needed ?
    # TODO: What are the valid keys ? (e.g. 'detector.image.array',
    #       'detector.signal.array' and 'detector.pixel.array')
    def get(self, key: str) -> Any:
        """Get the current value from a Parameter.

        Examples
        --------
        >>> processor = Processor()
        >>> processor.get("pipeline.photon_collection.illumination.arguments.level")
        array(0.)
        >>> processor.get("pipeline.characteristics.quantum_efficiency")
        array(0.1)
        """
        func: Callable = operator.attrgetter(key)
        result = func(self)

        return result

    # TODO: Could it be renamed '__setitem__' ?
    def set(  # noqa: A003
        self,
        key: str,
        value: Union[str, Number, np.ndarray, list[Union[str, Number, np.ndarray]]],
        convert_value: bool = True,
    ) -> None:
        """TBW.

        Parameters
        ----------
        key : str
        value
        convert_value : bool
        """
        if convert_value:  # and value:
            # TODO: Refactor this
            # convert the string based value to a number
            if isinstance(value, list):
                new_value_lst: list[Union[str, Number, np.ndarray]] = []

                val: Union[str, Number, np.ndarray]
                for val in value:
                    new_val = eval_entry(val) if val else val
                    new_value_lst.append(new_val)

                new_value: Union[
                    str, Number, np.ndarray, list[Union[str, Number, np.ndarray]]
                ] = new_value_lst

            else:
                converted_value: Union[str, Number, np.ndarray] = eval_entry(value)

                new_value = converted_value
        else:
            new_value = value

        obj, att = get_obj_att(self, key)

        if isinstance(obj, dict) and att in obj:
            obj[att] = new_value
        else:
            setattr(obj, att, new_value)

    # TODO: Create a method `DetectionPipeline.run`
    def run_pipeline(self, with_intermediate_steps: bool = False) -> None:
        """Run a pipeline with all its models in the right order.

        Parameters
        ----------
        with_intermediate_steps : bool

        Notes
        -----
        The ``Processor`` instance with method '.run_pipeline' is modified.
        """
        self._log.info("Start pipeline")

        for group_name in self.pipeline.model_group_names:
            # Get a group of models
            models_grp: Optional[ModelGroup] = getattr(self.pipeline, group_name)
            if not models_grp:
                continue

            self._log.info("Processing group: %r", group_name)
            models_grp.run(
                detector=self.detector,
                with_intermediate_steps=with_intermediate_steps,
            )

    # TODO: Refactor '.result'. See #524. Deprecate this method ?
    @property
    def result(self) -> dict:
        """Return exposure pipeline final result in a dictionary."""
        if not self._result:
            raise ValueError("No result saved in the processor.")
        return self._result

    # TODO: Refactor '.result'. See #524. Deprecate this method ?
    @result.setter
    def result(self, result_to_save: dict) -> None:
        """Set result."""
        self._result = result_to_save

    # TODO: Refactor '.result'. See #524
    def result_to_dataset(
        self,
        y: range,
        x: range,
        times: np.ndarray,
        result_type: ResultId,
    ) -> "xr.Dataset":
        """Return the result in a xarray dataset."""
        # Late import to speedup start-up time
        import xarray as xr

        if not self._result:
            raise ValueError("No result saved in the processor.")

        readout_time = xr.DataArray(
            times,
            dims=("readout_time",),
            attrs={"units": "s", "standard_name": "Readout time"},
        )

        lst: list[xr.DataArray] = []

        key: ResultId
        for key in result_keys(result_type):
            if key.startswith("data") or key.startswith("scene"):
                continue
            elif key == "photon":
                standard_name = "Photon"
                unit = "photon"
            elif key == "charge":
                standard_name = "Charge"
                unit = "electron"
            elif key == "pixel":
                standard_name = "Pixel"
                unit = "electron"
            elif key == "signal":
                standard_name = "Signal"
                unit = "volt"
            elif key == "image":
                standard_name = "Image"
                unit = "adu"
            else:
                raise NotImplementedError
                # standard_name = key
                # unit = ""

            da = xr.DataArray(
                self.result[key],
                dims=("readout_time", "y", "x"),
                name=key,
                coords={"readout_time": readout_time, "y": y, "x": x},
                attrs={"units": unit, "standard_name": standard_name},
            )

            lst.append(da)

        ds: xr.Dataset = xr.merge(lst, combine_attrs="drop_conflicts")
        ds.attrs.update({"pyxel version": __version__})

        return ds

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

import datatree.testing
import numpy as np
import xarray as xr
from datatree import DataTree

from pyxel.detectors import CCD, CCDGeometry, Characteristics, Environment
from pyxel.models.data_processing import mean_variance


def test_mean_variance():
    """Test model 'mean_variance'."""
    # Create fake photons for three different times
    rng = np.random.default_rng(seed=100)

    absolute_time = [0.0, 1.0, 2.0]
    photons = xr.DataArray(
        rng.random((3, 4, 5)),
        dims=["time", "y", "x"],
        coords={"time": absolute_time, "y": range(4), "x": range(5)},
    )

    # Created expected 'mean-variance'
    expected_variance = photons.var(dim=["y", "x"])
    expected_mean = photons.mean(dim=["y", "x"])

    expected_mean_variance = xr.Dataset()
    expected_mean_variance["variance"] = xr.DataArray(
        expected_variance.to_numpy(),
        dims=["mean"],
        coords={"mean": expected_mean.to_numpy()},
    )

    # Create fake CCD detector
    detector = CCD(
        geometry=CCDGeometry(row=4, col=5),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector.set_readout(times=[1.0, 2.0, 3.0])

    # Step 1
    detector.photon.array = photons.isel(time=0).to_numpy()
    mean_variance(detector, data_structure="photon")

    expected_photon_step1 = DataTree().from_dict(
        {"mean_variance/partial/photon": expected_mean_variance.isel(mean=[0])}
    )
    datatree.testing.assert_equal(detector.data, expected_photon_step1)

    # Step 2
    detector.photon.array = photons.isel(time=1).to_numpy()
    mean_variance(detector, data_structure="photon")

    expected_photon_step2 = DataTree().from_dict(
        {"mean_variance/partial/photon": expected_mean_variance.isel(mean=[1, 0])}
    )
    datatree.testing.assert_equal(detector.data, expected_photon_step2)

    # Step 3
    detector.photon.array = photons.isel(time=2).to_numpy()
    mean_variance(detector, data_structure="photon")

    expected_photon_step3 = DataTree().from_dict(
        {"mean_variance/partial/photon": expected_mean_variance.isel(mean=[2, 1, 0])}
    )
    datatree.testing.assert_equal(detector.data, expected_photon_step3)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


"""Single outputs."""

from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Protocol, Union

from pyxel.outputs import Outputs, ValidFormat, ValidName

if TYPE_CHECKING:
    import xarray as xr
    from datatree import DataTree

    class SaveToFile(Protocol):
        """TBW."""

        def __call__(self, data: Any, name: str, with_auto_suffix: bool = True) -> Path:
            """TBW."""
            ...


class ExposureOutputs(Outputs):
    """Collection of methods to save the data buckets from a Detector for an Exposure pipeline.

    Parameters
    ----------
    output_folder : str or Path
        Folder where sub-folder(s) that will be created to save data buckets.
    custom_dir_name : str, optional
        Prefix of the sub-folder name that will be created in the 'output_folder' folder.
        The default prefix is `run_`.
    save_data_to_file : Dict
        Dictionary where key is a 'data bucket' name (e.g. 'detector.photon.array') and value
        is the data format (e.g. 'fits').

        Example:
        {'detector.photon.array': 'fits', 'detector.charge.array': 'hdf', 'detector.image.array':'png'}
    """

    def __init__(
        self,
        output_folder: Union[str, Path],
        custom_dir_name: Optional[str] = "",
        save_data_to_file: Optional[
            Sequence[Mapping[ValidName, Sequence[ValidFormat]]]
        ] = None,
        save_exposure_data: Optional[Sequence[Mapping[str, Sequence[str]]]] = None,
    ):
        super().__init__(
            output_folder=output_folder,
            custom_dir_name=custom_dir_name,
            save_data_to_file=save_data_to_file,
        )

        self.save_exposure_data: Optional[
            Sequence[Mapping[str, Sequence[str]]]
        ] = save_exposure_data

    def save_exposure_outputs(self, dataset: Union["xr.Dataset", "DataTree"]) -> None:
        """Save the observation outputs such as the dataset.

        Parameters
        ----------
        dataset: Dataset
        """

        save_methods: dict[str, SaveToFile] = {"nc": self.save_to_netcdf}

        if self.save_exposure_data is None:
            return

        dct: Mapping[str, Sequence[str]]
        for dct in self.save_exposure_data:
            first_item, *_ = dct.items()
            obj, format_list = first_item

            if obj != "dataset":
                raise NotImplementedError(f"Object {obj} unknown.")

            out_format: str
            for out_format in format_list:
                if out_format not in save_methods:
                    raise ValueError(f"Format {out_format} not a valid save method!")

                func = save_methods[out_format]
                func(data=dataset, name=obj)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Model for replicating the gain register in an EMCCD."""

import numba
import numpy as np

from pyxel.detectors import CCD


def multiplication_register(
    detector: CCD,
    total_gain: int,
    gain_elements: int,
) -> None:
    """Calculate total gain of image with EMCCD multiplication register.

    Takes in CCD detector along with the gain and the total elements of the EMCCD
    multiplication register.

    Parameters
    ----------
    detector : CCD
    total_gain : int
    gain_elements : int
        Amount of single stage gain elements in the EMCCD register.
    """

    if total_gain < 0 or gain_elements < 0:
        raise ValueError("Wrong input parameter")

    detector.pixel.array = multiplication_register_poisson(
        image_cube=detector.pixel.array,
        total_gain=total_gain,
        gain_elements=gain_elements,
    ).astype(float)


@numba.njit
def poisson_register(lam, image_cube_pix, gain_elements):
    """Calculate the total gain of a single pixel from EMCCD register elements.

    A single pixel is
    inputted and is iterated through the total number of gain elements provided with the result being the resultant
    signal from the pixel going through the multiplication process.
    """

    new_image_cube_pix = image_cube_pix

    for _ in range(gain_elements):
        electron_gain = np.random.poisson(lam, size=int(new_image_cube_pix))
        new_image_cube_pix = np.round(image_cube_pix + np.sum(electron_gain))

    return new_image_cube_pix


@numba.njit
def multiplication_register_poisson(
    image_cube: np.ndarray,
    total_gain: int,
    gain_elements: int,
) -> np.ndarray:
    """Calculate total gain of image from EMCCD register.

    Cycles through each pixel within the image provided. Returns a final image with signal added.
    """

    new_image_cube = np.zeros_like(image_cube, dtype=np.int32)

    lam = total_gain ** (1 / gain_elements) - 1
    yshape, xshape = image_cube.shape

    for j in range(0, yshape):
        for i in range(0, xshape):
            if image_cube[j, i] < 0:
                new_image_cube[j, i] = poisson_register(
                    lam=lam, image_cube_pix=0, gain_elements=gain_elements
                )
            else:
                new_image_cube[j, i] = poisson_register(
                    lam=lam,
                    image_cube_pix=image_cube[j, i],
                    gain_elements=gain_elements,
                )

    return new_image_cube



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Model to compute linear regressions."""

from typing import TYPE_CHECKING, Literal, Optional, Union

import numpy as np

from pyxel.detectors import Detector

if TYPE_CHECKING:
    import xarray as xr

    from pyxel.data_structure import Image, Photon, Pixel, Signal


def linear_regression(
    detector: Detector,
    data_structure: Literal["pixel", "photon", "image", "signal"] = "image",
    name: Optional[str] = None,
) -> None:
    """Compute a linear regression along 'readout_time' and store it in '.data' bucket.

    Parameters
    ----------
    detector : Detector
    data_structure : 'pixel', 'photon', 'image' or 'signal'
        Data bucket to use for the linear regression.
    name : str, optional
        Name to use for the result.

    Raises
    ------
    ValueError
        If less than 2 timing steps is used.

    Examples
    --------
    >>> import pyxel
    >>> config = pyxel.load("exposure_mode.yaml")
    >>> linear_regression(detector=detector)

    Run exposure mode with 'data_processing/linear_regression' model

    >>> data_tree = pyxel.run_mode(
    ...     mode=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )

    Get results

    >>> data_tree["/data/linear_regression"]
    DataTree('linear_regression', parent="data")
    └── DataTree('image')
            Dimensions:        (y: 100, x: 100)
            Coordinates:
              * y              (y) int64 0 1 2 3 4 5 6 7 8 9 ... 91 92 93 94 95 96 97 98 99
              * x              (x) int64 0 1 2 3 4 5 6 7 8 9 ... 91 92 93 94 95 96 97 98 99
            Data variables:
                slope          (y, x) float64 396.8 396.2 396.7 397.0 ... 396.1 396.2 396.7
                intercept      (y, x) float64 8.363e+03 8.356e+03 ... 8.364e+03 8.36e+03
                r2             (y, x) float64 0.9313 0.9312 0.9316 ... 0.9309 0.931 0.9313
                slope_std      (y, x) float64 14.94 14.93 14.91 14.91 ... 14.96 14.96 14.94
                intercept_std  (y, x) float64 846.4 845.8 844.5 844.5 ... 847.7 847.6 846.1

    Display slope as a 2D map

    >>> data_tree["/data/linear_regression/image/slope"].plot(robust=True)

    .. figure:: _static/linear_regression_slope.png
        :scale: 70%
        :alt: Linear Regression Slope
        :align: center
    """
    if detector.num_steps < 3:
        raise ValueError(f"Expecting at least 2 steps. Got: {detector.num_steps}")

    if name is None:
        name = data_structure

    # Extract data from 'detector'
    data_bucket: Union[Pixel, Photon, Image, Signal] = getattr(detector, data_structure)

    x: float = detector.absolute_time
    y: xr.DataArray = data_bucket.to_xarray(dtype=float)

    if detector.is_first_readout:
        detector.data[f"/linear_regression/{name}/partial/n"] = 1
        detector.data[f"/linear_regression/{name}/partial/sum_x"] = x
        detector.data[f"/linear_regression/{name}/partial/sum_y"] = y
        detector.data[f"/linear_regression/{name}/partial/sum_xy"] = x * y
        detector.data[f"/linear_regression/{name}/partial/sum_x2"] = x * x
        detector.data[f"/linear_regression/{name}/partial/sum_y2"] = y * y
    else:
        detector.data[f"/linear_regression/{name}/partial/n"] += 1  # type: ignore
        detector.data[f"/linear_regression/{name}/partial/sum_x"] += x  # type: ignore
        detector.data[f"/linear_regression/{name}/partial/sum_y"] += y
        detector.data[f"/linear_regression/{name}/partial/sum_xy"] += x * y
        detector.data[f"/linear_regression/{name}/partial/sum_x2"] += x * x  # type: ignore
        detector.data[f"/linear_regression/{name}/partial/sum_y2"] += y * y

        if detector.is_last_readout:
            n = detector.data[f"/linear_regression/{name}/partial/n"]
            sum_x = detector.data[f"/linear_regression/{name}/partial/sum_x"]
            sum_y = detector.data[f"/linear_regression/{name}/partial/sum_y"]
            sum_xy = detector.data[f"/linear_regression/{name}/partial/sum_xy"]
            sum_x2 = detector.data[f"/linear_regression/{name}/partial/sum_x2"]
            sum_y2 = detector.data[f"/linear_regression/{name}/partial/sum_y2"]

            factor_x = n * sum_x2 - sum_x * sum_x  # type: ignore
            factor_y = n * sum_y2 - sum_y * sum_y  # type: ignore

            slope = (n * sum_xy - sum_x * sum_y) / factor_x  # type: ignore
            intercept = (sum_y - slope * sum_x) / n

            r2 = slope * slope * factor_x / factor_y

            factor_var = factor_y - slope * slope * factor_x
            slope_var = factor_var / ((n - 2) * factor_x)  # type: ignore
            intercept_var = sum_x2 * factor_var / (n * (n - 2) * factor_x)  # type: ignore

            slope_std = np.sqrt(slope_var)
            intercept_std = np.sqrt(intercept_var)

            if y_unit := y.attrs.get("units"):
                x_unit = "s"
                slope.attrs["unit"] = f"{y_unit}/{x_unit}"
                intercept.attrs["unit"] = y_unit
                slope_std.attrs["unit"] = f"{y_unit}/{x_unit}"  # type: ignore
                intercept_std.attrs["unit"] = y_unit  # type: ignore

            detector.data[f"/linear_regression/{name}/slope"] = slope
            detector.data[f"/linear_regression/{name}/intercept"] = intercept
            detector.data[f"/linear_regression/{name}/r2"] = r2
            detector.data[f"/linear_regression/{name}/slope_std"] = slope_std
            detector.data[f"/linear_regression/{name}/intercept_std"] = intercept_std

            detector.data[f"/linear_regression/{name}"].attrs[
                "long_name"
            ] = f"Linear regression: {name}"

            # Remove '/linear_regression/partial
            detector.data[f"/linear_regression/{name}/partial"].orphan()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from pathlib import Path
from typing import Literal, Optional

import numpy as np
import pytest

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Environment,
    ReadoutProperties,
)
from pyxel.models.photon_collection import load_image


@pytest.fixture
def valid_data2d(
    tmp_path: Path,
) -> Path:
    """Create valid 2D file on a temporary folder."""
    data_2d = np.ones((20, 20), dtype=np.uint16)
    final_path = f"{tmp_path}/img.npy"
    np.save(final_path, arr=data_2d)

    data_2d = np.ones((5, 5), dtype=np.uint16)
    final_path = f"{tmp_path}/img_invalid.npy"
    np.save(final_path, arr=data_2d)

    return tmp_path


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(
            quantum_efficiency=1.0,
            charge_to_volt_conversion=1.0e-6,
            pre_amplification=1.0,
            adc_bit_resolution=16,
            adc_voltage_range=(0.0, 10.0),
        ),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.mark.parametrize(
    "image_file, position, align, convert_to_photons, multiplier, time_scale, bit_resolution",
    [
        pytest.param("img.npy", (0, 0), None, True, 1.0, 1.0, 16, id="valid"),
    ],
)
def test_load_image(
    ccd_10x10: CCD,
    valid_data2d: str,
    image_file: str,
    position: tuple[int, int],
    align: Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ],
    convert_to_photons: bool,
    multiplier: float,
    time_scale: float,
    bit_resolution: int,
):
    """Test input parameters for function 'load_image'."""
    load_image(
        detector=ccd_10x10,
        image_file=f"{valid_data2d}/{image_file}",
        position=position,
        align=align,
        convert_to_photons=convert_to_photons,
        multiplier=multiplier,
        time_scale=time_scale,
        bit_resolution=bit_resolution,
    )


@pytest.mark.parametrize(
    "image_file, position, align, convert_to_photons, multiplier, time_scale, bit_resolution, exp_exc, exp_msg",
    [
        pytest.param(
            "img.npy",
            (0, 0),
            None,
            True,
            1.0,
            1.0,
            None,
            ValueError,
            "Bit resolution of the input image has to be specified for converting to photons.",
            id="invalid",
        ),
    ],
)
def test_load_image_with_invalid_params(
    ccd_10x10: CCD,
    valid_data2d: str,
    image_file: str,
    position: tuple[int, int],
    align: Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ],
    convert_to_photons: bool,
    multiplier: float,
    time_scale: float,
    bit_resolution: int,
    exp_exc,
    exp_msg,
):
    """Test input parameters for function 'load_image'."""
    with pytest.raises(exp_exc, match=exp_msg):
        load_image(
            detector=ccd_10x10,
            image_file=f"{valid_data2d}/{image_file}",
            position=position,
            align=align,
            convert_to_photons=convert_to_photons,
            multiplier=multiplier,
            time_scale=time_scale,
            bit_resolution=bit_resolution,
        )



import pytest

from pyxel.detectors import CCD
from pyxel.models.charge_transfer.emccd_poisson import multiplication_register


def test_multiplication_register(ccd_10x10: CCD):
    detector = ccd_10x10

    multiplication_register(detector=detector, total_gain=0.0, gain_elements=1)


@pytest.mark.parametrize(
    "total_gain,gain_elements",
    [
        (-1, 10),
        (1, -1),
        (-1, -1),
    ],
)
def test_multiplication_register_bad_inputs(ccd_10x10: CCD, total_gain, gain_elements):
    detector = ccd_10x10

    with pytest.raises(ValueError, match="Wrong input parameter"):
        multiplication_register(
            detector=detector, total_gain=total_gain, gain_elements=gain_elements
        )



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

"""Models to generate charge due to dark current process."""
import warnings
from typing import Optional

import numpy as np
from astropy import constants as const

from pyxel.detectors import APD, Detector
from pyxel.util import set_random_seed


def calculate_band_gap(
    band_gap_0: float, alpha: float, beta: float, temperature: float
) -> float:
    """Return band gap based on Varshni empirical expression.

    Parameters
    ----------
    band_gap_0 : float
        Parameter E_0. Unit: eV
    alpha : float
        Alpha parameter. Unit: eV/K
    beta : float
        Beta parameter. Unit: K
    temperature :
        Temperature. Unit K

    Returns
    -------
    float
        Band gap value. Unit: eV
    """

    gap = band_gap_0 - (alpha * temperature**2) / (temperature + beta)

    return gap


def band_gap_silicon(temperature: float) -> float:
    """Return band gap in Silicon based on Varshni empirical expression.

    Parameters
    ----------
    temperature : float
        Temperature. Unit: K

    Returns
    -------
    float
        Band gap value. Unit: eV
    """

    band_gap_0 = 1.1557  # eV
    alpha = 7.021e-4  # ev/K
    beta = 1108.0  # K

    return calculate_band_gap(
        band_gap_0=band_gap_0, alpha=alpha, beta=beta, temperature=temperature
    )


def average_dark_current(
    temperature: float,
    pixel_area: float,
    figure_of_merit: float,
    band_gap: float,
    band_gap_room_temperature: float,
) -> float:
    """Compute average dark current.

    Parameters
    ----------
    temperature : float
        Temperature. Unit: K
    pixel_area
        Pixel area. Unit: cm^2
    figure_of_merit : float
        Dark current figure of merit. Unit: nA/cm^2
    band_gap : float
        Semiconductor band_gap. Unit: eV
    band_gap_room_temperature : float
        Semiconductor band gap at 300K. If none, the one for silicon is used. Unit: eV

    Returns
    -------
    float
        Average dark current. Unit: e-/pixel/s
    """

    k_b = const.k_B.value
    e_0 = const.e.value

    room_temperature = 300

    room_temperature_factor = room_temperature ** (3 / 2) * np.exp(
        -band_gap_room_temperature * e_0 / (2 * k_b * room_temperature)
    )

    avg_dark_current = (
        pixel_area  # in cm^2
        * figure_of_merit  # in nA/cm^2
        * 1e-9  # conversion to A/cm^2
        * (1 / e_0)  # conversion to e-/s/cm^2
        * temperature ** (3 / 2)
        * np.exp(-band_gap * e_0 / (2 * k_b * temperature))
        * (1 / room_temperature_factor)
    )  # Unit: e-/s/pixel

    return avg_dark_current


def compute_dark_current(
    shape: tuple[int, int],
    time_step: float,
    temperature: float,
    pixel_area: float,
    figure_of_merit: float,
    band_gap: float,
    band_gap_room_temperature: float,
    spatial_noise_factor: Optional[float] = None,
    temporal_noise: bool = True,
) -> np.ndarray:
    """Compute dark current.

    Based on:
    Konnik, Mikhail V. and James S. Welsh.
    “High-level numerical simulations of noise in CCD and CMOS photosensors: review and tutorial.”
    ArXiv abs/1412.4031 (2014): n. pag.

    Parameters
    ----------
    shape : tuple
        Output array shape.
    time_step : float
        Time step. Unit: s
    temperature : float
        Temperature. Unit: K
    pixel_area
        Pixel area. Unit: cm^2
    figure_of_merit : float
        Dark current figure of merit. Unit: nA/cm^2
    band_gap : float
        Semiconductor band_gap. Unit: eV
    band_gap_room_temperature : float
        Semiconductor band gap at 300K. If none, the one for silicon is used. Unit: eV
    spatial_noise_factor: float
        Dark current fixed pattern noise factor.
    temporal_noise: bool
        Shot noise.

    Returns
    -------
    ndarray
        Dark current values. Unit: e-
    """

    avg_dark_current = average_dark_current(
        temperature=temperature,
        pixel_area=pixel_area,
        figure_of_merit=figure_of_merit,
        band_gap=band_gap,
        band_gap_room_temperature=band_gap_room_temperature,
    )

    if temporal_noise:
        dark_signal_2d = np.ones(shape) * avg_dark_current * time_step
        dark_current_shot_noise_2d = np.random.poisson(
            dark_signal_2d
        )  # dark current shot noise
        dark_current_2d = dark_current_shot_noise_2d.astype(float)
    else:
        dark_signal_2d = np.ones(shape) * avg_dark_current * time_step
        dark_current_2d = dark_signal_2d

    if spatial_noise_factor is not None:
        dark_current_fpn_sigma = (
            time_step * avg_dark_current * spatial_noise_factor
        )  # sigma of fpn distribution

        dark_current_2d = dark_current_2d * (
            1 + np.random.lognormal(sigma=dark_current_fpn_sigma, size=shape)
        )

    if np.isinf(dark_current_2d).any():
        warnings.warn(
            "Unphysical high value for dark current from fixed pattern noise distribution"
            " will result in inf values. Enable a FWC model to ensure a physical limit.",
            RuntimeWarning,
            stacklevel=2,
        )

    return dark_current_2d


def dark_current(
    detector: Detector,
    figure_of_merit: float,
    spatial_noise_factor: Optional[float] = None,
    band_gap: Optional[float] = None,
    band_gap_room_temperature: Optional[float] = None,
    seed: Optional[int] = None,
    temporal_noise: bool = True,
) -> None:
    """Add dark current to the detector charge.

    Based on:
    Konnik, Mikhail V. and James S. Welsh.
    “High-level numerical simulations of noise in CCD and CMOS photosensors: review and tutorial.”
    ArXiv abs/1412.4031 (2014): n. pag.

    Parameters
    ----------
    detector : Detector
        Pyxel detector object.
    figure_of_merit : float
        Dark current figure of merit. Unit: nA/cm^2
    spatial_noise_factor : float
        Dark current fixed pattern noise factor.
    band_gap : float, optional
        Semiconductor band_gap. If none, the one for silicon is used. Unit: eV
    band_gap_room_temperature : float, optional
        Semiconductor band gap at 300K. If none, the one for silicon is used. Unit: eV
    seed : int, optional
        Random seed.
    temporal_noise : bool, optional
        Shot noise.
    """
    geo = detector.geometry
    pixel_area = geo.pixel_vert_size * 1e-4 * geo.pixel_horz_size * 1e-4  # in cm^2
    temperature = detector.environment.temperature
    time_step = detector.time_step

    if band_gap and band_gap_room_temperature:
        final_band_gap = band_gap
        final_band_gap_room_temperature = band_gap_room_temperature
    elif band_gap or band_gap_room_temperature:
        raise ValueError(
            "Both parameters band_gap and band_gap_room_temperature have to be defined."
        )
    else:
        final_band_gap = band_gap_silicon(temperature)
        final_band_gap_room_temperature = band_gap_silicon(temperature=300)

    with set_random_seed(seed):
        dark_current_array = compute_dark_current(
            shape=geo.shape,
            time_step=time_step,
            temperature=temperature,
            pixel_area=pixel_area,
            figure_of_merit=figure_of_merit,
            band_gap=final_band_gap,
            band_gap_room_temperature=final_band_gap_room_temperature,
            spatial_noise_factor=spatial_noise_factor,
            temporal_noise=temporal_noise,
        )

    detector.charge.add_charge_array(dark_current_array)


def calculate_simple_dark_current(
    num_rows: int, num_cols: int, current: float, exposure_time: float
) -> np.ndarray:
    """Simulate dark current in a :term:`CCD`.

    Parameters
    ----------
    num_rows : int
        Number of rows for the generated image.
    num_cols : int
        Number of columns for the generated image.
    current : float
        Dark current, in electrons/pixel/second
    exposure_time : float
        Length of the simulated exposure, in seconds.

    Returns
    -------
    ndarray
        An array the same shape and dtype as the input containing dark counts
        in units of charge (e-).
    """
    # dark current for every pixel
    mean_dark_charge = current * exposure_time

    # This random number generation should change on each call.
    dark_im_array_2d = np.random.poisson(mean_dark_charge, size=(num_rows, num_cols))
    return dark_im_array_2d


def simple_dark_current(
    detector: Detector, dark_rate: float, seed: Optional[int] = None
) -> None:
    """Simulate dark current in a detector.

    Parameters
    ----------
    detector : Detector
        Any detector object.
    dark_rate : float
        Dark current, in electrons/pixel/second, which is the way
        manufacturers typically report it.
    seed : int, optional
    """

    exposure_time = detector.time_step
    geo = detector.geometry

    with set_random_seed(seed):
        dark_current_array: np.ndarray = calculate_simple_dark_current(
            num_rows=geo.row,
            num_cols=geo.col,
            current=dark_rate,
            exposure_time=exposure_time,
        ).astype(float)

    detector.charge.add_charge_array(dark_current_array)


def calculate_dark_current_saphira(
    temperature: float,
    avalanche_gain: float,
    shape: tuple[int, int],
    exposure_time: float,
) -> np.ndarray:
    """Simulate dark current in a Saphira :term:`APD`.

    From: I. M. Baker et al., Linear-mode avalanche photodiode arrays in HgCdTe at Leonardo, UK: the
    current status, in Image Sensing Technologies: Materials, Devices, Systems, and Applications VI,
    2019, vol. 10980, no. May, p. 20.

    Parameters
    ----------
    temperature
    avalanche_gain
    shape
    exposure_time : float
        Length of the simulated exposure, in seconds.

    Returns
    -------
    ndarray
        An array the same shape and dtype as the input containing dark counts
        in units of charge (e-).
    """

    # We can split the dark current vs. gain vs. temp plot ([5] Fig. 3) into three linear
    # 'regimes': 1) low-gain, low dark current; 2) nominal; and 3) trap-assisted tunneling.
    # The following ignores (1) for now since this only applies at gains less than ~2.

    dark_nominal = (15.6e-12 * np.exp(0.2996 * temperature)) * (
        avalanche_gain
        ** ((-5.43e-4 * (temperature**2)) + (0.0669 * temperature) - 1.1577)
    )
    dark_tunnel = 3e-5 * (avalanche_gain**3.2896)

    # The value we want is the maximum of the two regimes (see the plot):
    dark = max(dark_nominal, dark_tunnel)

    # dark current for every pixel
    mean_dark_charge = dark * exposure_time

    # This random number generation should change on each call.
    dark_im_array_2d = np.random.poisson(mean_dark_charge, size=shape)

    return dark_im_array_2d


def dark_current_saphira(detector: APD, seed: Optional[int] = None) -> None:
    """Simulate dark current in a Saphira APD detector.

    Reference: I. M. Baker et al., Linear-mode avalanche photodiode arrays in HgCdTe at Leonardo, UK: the
    current status, in Image Sensing Technologies: Materials, Devices, Systems, and Applications VI,
    2019, vol. 10980, no. May, p. 20.

    Parameters
    ----------
    detector : APD
        An APD detector object.
    seed : int, optional
    """

    if not isinstance(detector, APD):
        raise TypeError("Expecting an APD object for detector.")
    if detector.environment.temperature > 100:
        raise ValueError(
            "Dark current estimation is inaccurate for temperatures more than 100 K!"
        )
    if detector.characteristics.avalanche_gain < 2:
        raise ValueError("Dark current is inaccurate for avalanche gains less than 2!")

    exposure_time = detector.time_step

    with set_random_seed(seed):
        dark_current_array: np.ndarray = calculate_dark_current_saphira(
            temperature=detector.environment.temperature,
            avalanche_gain=detector.characteristics.avalanche_gain,
            shape=detector.geometry.shape,
            exposure_time=exposure_time,
        ).astype(float)

    detector.charge.add_charge_array(dark_current_array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
import sys

import astropy.constants as const
import astropy.units as u
import numpy as np
import pytest
import xarray as xr
from astropy.table import Table
from astropy.tests.helper import assert_quantity_allclose
from datatree import DataTree
from pytest_mock import MockerFixture  # pip install pytest-mock

from pyxel.detectors import CCD
from pyxel.models.scene_generation import load_star_map
from pyxel.models.scene_generation.load_star_map import GaiaPassBand, retrieve_from_gaia


@pytest.fixture
def source_ids() -> list[int]:
    """Return unique source identifier from the GAIA database."""
    return [66727234683960320, 65214031805717376, 65225851555715328, 65226195153096192]


@pytest.fixture
def positions(source_ids: list[int]) -> xr.Dataset:
    """Return source objects as a Dataset."""
    source_id = xr.DataArray(
        source_ids,
        dims="source_id",
    )

    ds = xr.Dataset(coords={"source_id": source_id})
    ds["ra"] = xr.DataArray(
        [56.760485086776846, 56.74561610052, 56.726951308177455, 56.736700233543914],
        dims="source_id",
        attrs={"units": "deg"},
    )

    ds["dec"] = xr.DataArray(
        [24.149991010998, 24.089174782613, 24.111718134110, 24.149504345515],
        dims="source_id",
        attrs={"units": "deg"},
    )
    ds["has_xp_sampled"] = xr.DataArray([True] * 4, dims="source_id")
    ds["phot_bp_mean_mag"] = xr.DataArray(
        [14.734505, 12.338661, 14.627676, 14.272486],
        dims="source_id",
        attrs={"units": "mag"},
    )
    ds["phot_g_mean_mag"] = xr.DataArray(
        [14.433147, 11.940813, 13.91212, 13.613182],
        dims="source_id",
        attrs={"units": "mag"},
    )
    ds["phot_rp_mean_mag"] = xr.DataArray(
        [13.954827, 11.368548, 13.091035, 12.804853],
        dims="source_id",
        attrs={"units": "mag"},
    )

    return ds


@pytest.fixture
def positions_table(positions: xr.Dataset) -> Table:
    """Return source objects as an Astropy Table."""
    table = Table.from_pandas(positions.to_pandas().reset_index())
    table["ra"].unit = u.deg
    table["dec"].unit = u.deg
    table["phot_bp_mean_mag"].unit = u.mag
    table["phot_g_mean_mag"].unit = u.mag
    table["phot_rp_mean_mag"].unit = u.mag

    return table


@pytest.fixture
def wavelengths() -> xr.DataArray:
    """Return wavelengths."""
    return xr.DataArray(
        [336.0, 338.0, 1018.0, 1020.0],
        dims="wavelength",
        attrs={"dims": "nm"},
    )


@pytest.fixture
def spectra1(wavelengths: xr.DataArray) -> xr.Dataset:
    """Return spectra for the first object."""
    ds = xr.Dataset(coords={"wavelength": wavelengths})
    ds["flux"] = xr.DataArray(
        [4.1858373e-17, 4.1012171e-17, 1.3888942e-17, 1.3445790e-17],
        dims="wavelength",
        attrs={"dims": "W / (nm * m^2)"},
    )
    ds["flux_error"] = xr.DataArray(
        [5.8010027e-18, 4.3436358e-18, 4.2265315e-18, 4.1775913e-18],
        dims="wavelength",
        attrs={"dims": "W / (nm * m^2)"},
    )

    return ds


@pytest.fixture
def spectra2(wavelengths: xr.DataArray) -> xr.Dataset:
    """Return spectra for the second object."""
    ds = xr.Dataset(coords={"wavelength": wavelengths})
    ds["flux"] = xr.DataArray(
        [3.0237057e-16, 2.9785625e-16, 2.4918341e-16, 2.5573007e-16],
        dims="wavelength",
        attrs={"dims": "W / (nm * m^2)"},
    )
    ds["flux_error"] = xr.DataArray(
        [2.8953863e-17, 2.1999507e-17, 1.7578710e-17, 1.7167379e-17],
        dims="wavelength",
        attrs={"dims": "W / (nm * m^2)"},
    )

    return ds


@pytest.fixture
def spectra3(wavelengths: xr.DataArray) -> xr.Dataset:
    """Return spectra for the third object."""
    ds = xr.Dataset(coords={"wavelength": wavelengths})
    ds["flux"] = xr.DataArray(
        [2.0389929e-17, 2.2613652e-17, 6.1739404e-17, 6.5074511e-17],
        dims="wavelength",
        attrs={"dims": "W / (nm * m^2)"},
    )
    ds["flux_error"] = xr.DataArray(
        [6.6810894e-18, 5.1004213e-18, 7.3332771e-18, 7.2768021e-18],
        dims="wavelength",
        attrs={"dims": "W / (nm * m^2)"},
    )

    return ds


@pytest.fixture
def spectra4(wavelengths: xr.DataArray) -> xr.Dataset:
    """Return spectra for the fourth object."""
    ds = xr.Dataset(coords={"wavelength": wavelengths})
    ds["flux"] = xr.DataArray(
        [2.4765946e-17, 2.1556272e-17, 9.0950110e-17, 9.1888827e-17],
        dims="wavelength",
        attrs={"dims": "W / (nm * m^2)"},
    )
    ds["flux_error"] = xr.DataArray(
        [6.9233657e-18, 5.2693949e-18, 7.4365125e-18, 7.1868190e-18],
        dims="wavelength",
        attrs={"dims": "W / (nm * m^2)"},
    )

    return ds


@pytest.fixture
def spectra_dct(
    source_ids: list[int],
    spectra1: xr.Dataset,
    spectra2: xr.Dataset,
    spectra3: xr.Dataset,
    spectra4: xr.Dataset,
) -> dict[int, Table]:
    """Return spectra for the four objects as dictionary."""
    dct = {}
    for source_id, spectra in zip(source_ids, [spectra1, spectra2, spectra3, spectra4]):
        table: Table = Table.from_pandas(spectra.to_pandas().reset_index())
        table["wavelength"].unit = u.nm
        table["flux"].unit = u.W / (u.nm * u.m * u.m)
        table["flux_error"].unit = u.W / (u.nm * u.m * u.m)

        dct[source_id] = table

    return dct


@pytest.fixture
def source1_gaia(
    positions: xr.Dataset,
    source_ids: list[int],
    spectra1: xr.Dataset,
    spectra2: xr.Dataset,
    spectra3: xr.Dataset,
    spectra4: xr.Dataset,
) -> xr.Dataset:
    """Build a Dataset object retrieved from the Gaia database."""
    return xr.merge(
        [
            positions,
            xr.concat(
                [
                    spectra.assign(source_id=source_id)
                    for source_id, spectra in zip(
                        source_ids, [spectra1, spectra2, spectra3, spectra4]
                    )
                ],
                dim="source_id",
            ),
        ]
    )


@pytest.fixture
def source1_pyxel(
    positions: xr.Dataset,
    source_ids: list[int],
    spectra1: xr.Dataset,
    spectra2: xr.Dataset,
    spectra3: xr.Dataset,
    spectra4: xr.Dataset,
) -> DataTree:
    return 0


@pytest.mark.skipif(sys.version_info < (3, 11), reason="Requires Python 3.11+")
def test_retrieve_from_gaia(
    mocker: MockerFixture,
    positions_table: Table,
    spectra_dct: dict[int, Table],
    source1_pyxel: xr.Dataset,
    source1_gaia: xr.Dataset,
):
    """Test function 'retrieve_from_gaia'."""
    # Mock function 'pyxel.models.scene_generation.load_star_map.retrieve_objects_from_gaia'
    # When this function will be called (with any parameters), it will always return this tuple
    # (positions_table, spectra_dct)
    mocker.patch(
        target="pyxel.models.scene_generation.load_star_map._retrieve_objects_from_gaia",
        return_value=(positions_table, spectra_dct),
    )

    ds = retrieve_from_gaia(
        right_ascension=0.0,  # This parameter is not important
        declination=0.0,  # This parameter is not important
        fov_radius=0.0,  # This parameter is not important
    )

    expected_ds = source1_gaia
    xr.testing.assert_equal(ds, expected_ds)


@pytest.mark.parametrize(
    "name, exp_band",
    [
        ("blue_photometer", GaiaPassBand.BluePhotometer),
        ("gaia_band", GaiaPassBand.GaiaBand),
        ("red_photometer", GaiaPassBand.RedPhotometer),
    ],
)
def test_gaia_passband(name: str, exp_band: GaiaPassBand):
    """Test class 'GaiaPassBand'."""
    band = GaiaPassBand(name)
    assert band is exp_band


@pytest.mark.parametrize(
    "band, exp_result", [(GaiaPassBand.BluePhotometer, "phot_bp_mean_mag")]
)
def test_gaia_pass_band_magnitude_key(band: GaiaPassBand, exp_result: str):
    """Test method 'GaiaPassBand.get_magnitude_key()."""
    assert isinstance(band, GaiaPassBand)

    result = band.get_magnitude_key()
    assert result == exp_result


def test_compute_flux_compare_to_astropy():
    """Test function 'compute_flux'."""
    wavelengths = u.Quantity([336.0, 338.0, 1018.0, 1020.0], unit=u.nm)
    flux = u.Quantity(
        [4.1858373e-17, 4.1012171e-17, 1.3888942e-17, 1.3445790e-17],
        unit="W / (nm * m2)",
    )
    expected_flux = u.Quantity(
        [0.00070802, 0.00069783, 0.00071177, 0.00069041],
        unit="ph / (Angstrom s cm2)",
    )

    # Compute new flux using astropy
    new_flux = [
        y.to(u.Unit("ph / (Angstrom s cm2)"), equivalencies=u.spectral_density(x))
        for x, y in zip(wavelengths, flux)
    ]

    assert_quantity_allclose(actual=new_flux, desired=expected_flux, rtol=1e-5)


def test_compute_flux_compare_to_manual_conversion():
    """Test function 'compute_flux'."""
    wavelengths = u.Quantity([336.0, 338.0, 1018.0, 1020.0], unit=u.nm)
    flux = u.Quantity(
        [4.1858373e-17, 4.1012171e-17, 1.3888942e-17, 1.3445790e-17],
        unit="W / (nm * m2)",
    )
    expected_flux = u.Quantity(
        [0.00070802, 0.00069783, 0.00071177, 0.00069041],
        unit="ph / (Angstrom s cm2)",
    )

    # Compute new flux using the Plank's equation 'E = h * v = h * (c / wavelength)'
    photon_energy = (const.h * const.c / wavelengths).to(u.J) / u.ph
    new_flux = (flux / photon_energy).to("ph / (Angstrom s cm2)")

    assert_quantity_allclose(actual=new_flux, desired=expected_flux, rtol=1e-5)


@pytest.mark.skipif(sys.version_info < (3, 11), reason="Requires Python 3.11+")
def test_load_star_map(
    mocker: MockerFixture,
    ccd_10x10: CCD,
    positions_table: Table,
    spectra_dct: dict[int, Table],
    wavelengths: xr.DataArray,
    source1_gaia,
    source1_pyxel: DataTree,
):
    """Test model 'load_star_map'."""

    # Mock function 'pyxel.models.scene_generation.load_star_map.retrieve_objects_from_gaia'
    # When this function will be called (with any parameters), it will always return this tuple
    # (positions_table, spectra_dct)
    mocker.patch(
        target="pyxel.models.scene_generation.load_star_map._retrieve_objects_from_gaia",
        return_value=(positions_table, spectra_dct),
    )

    # Run model
    detector: CCD = ccd_10x10
    load_star_map(
        detector=detector,
        right_ascension=0.0,  # This parameter is not important
        declination=0.0,  # This parameter is not important
        fov_radius=0.0,  # This parameter is not important
    )

    # Check outputs
    scene = detector.scene.data
    assert isinstance(scene, DataTree)

    expected_x = 3600 * source1_gaia["ra"]
    expected_y = 3600 * source1_gaia["dec"]

    exp_ds = xr.Dataset()
    exp_ds["x"] = xr.DataArray(
        np.array(expected_x), dims="ref", coords={"ref": [0, 1, 2, 3]}
    )
    exp_ds["y"] = xr.DataArray(
        np.array(expected_y), dims="ref", coords={"ref": [0, 1, 2, 3]}
    )
    exp_ds["weight"] = xr.DataArray(
        np.array(source1_gaia["phot_bp_mean_mag"]),
        dims="ref",
        coords={"ref": [0, 1, 2, 3]},
    )

    x = u.Quantity(wavelengths, unit="nm")
    flux = u.Quantity(source1_gaia["flux"], unit="W / (nm * m2)")

    # Compute new flux using the Plank's equation 'E = h * v = h * (c / wavelength)'
    photon_energy = (const.h * const.c / x).to(u.J) / u.ph
    new_flux = (flux / photon_energy).to("ph / (nm s cm2)")

    flux_converted: xr.DataArray = xr.zeros_like(scene["/list/0"]["flux"])
    flux_converted[:] = new_flux
    exp_ds["flux"] = flux_converted

    assert "list" in scene
    assert "0" in scene["list"]
    xr.testing.assert_allclose(scene["/list/0"].to_dataset(), exp_ds)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

""":term:`SAR` :term:`ADC` model with noise."""

import numpy as np

from pyxel.detectors import Detector


def apply_sar_adc_with_noise(
    signal_2d: np.ndarray,
    num_rows: int,
    num_cols: int,
    strengths: np.ndarray,
    noises: np.ndarray,
    max_volt: float,
    adc_bits: int,
) -> np.ndarray:
    """Apply :term:`SAR` :term:`ADC` with noise.

    Parameters
    ----------
    signal_2d : ndarray
    num_rows : int
    num_cols : int
    strengths : ndarray
    noises : ndarray
    max_volt : float
        Max volt of the ADC.
    adc_bits : int
        Number of bits the value will be encoded with.

    Returns
    -------
    ndarray
        2D digitized array.
    """
    data_digitized_2d = np.zeros((num_rows, num_cols))

    signal_normalized_2d = signal_2d.copy()

    # Set the reference voltage of the ADC to half the max
    ref_2d = np.full(shape=(num_rows, num_cols), fill_value=max_volt / 2.0)

    # For each bits, compare the value of the ref to the capacitance value
    for i in np.arange(adc_bits):
        strength = strengths[i]
        noise = noises[i]

        # digital value associated with this step
        digital_value: int = 2 ** (adc_bits - (i + 1))

        ref_2d += np.random.normal(loc=strength, scale=noise, size=(num_rows, num_cols))

        # All data that is higher than the ref is equal to the dig. value
        mask_2d: np.ndarray = signal_normalized_2d >= ref_2d
        data_digitized_2d += digital_value * mask_2d

        # Subtract ref value from the data
        signal_normalized_2d -= ref_2d * mask_2d

        # Divide reference voltage by 2 for next step
        ref_2d /= 2.0

    return data_digitized_2d


# TODO: documentation, range volt - only max is used
def sar_adc_with_noise(
    detector: Detector,
    strengths: tuple[float, ...],
    noises: tuple[float, ...],
) -> None:
    r"""Digitize signal array using :term:`SAR` (Successive Approximation Register) :term:`ADC` logic with noise.

    Successive-approximation-register (:term:`SAR`) analog-to-digital converters (:term:`ADC`)
    for each bit, will compare the randomly perturbated **reference voltage** to
    the voltage of the pixel, adding the corresponding digital value if it is superior.
    The perturbations are generated randomly for each pixel.

    The reference voltage fluctuations are regulated by two parameters, **strength** and
    **noise**, that are set independently for each bit.

    This model and the notes are based on the work of Emma Esparza Borges
    from :term:`ESO`.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    strengths : tuple of float
        Sequence of ``detector.characteristics.adc_bit_resolution`` number(s). Unit: V
    noises : tuple of float
        Sequence of ``detector.characteristics.adc_bit_resolution`` number(s). Unit: V

    Raises
    ------
    ValueError
        Raised if parameters ``strengths`` and/or ``noises`` does not have exactly the
        correct numbers of parameters.
        It is expected to have ``detector.characteristics.adc_bit_resolution`` elements.

    Notes
    -----
    The :term:`SAR` :term:`ADC` architecture is based on a binary search algorithm.
    Essentially, it works comparing the analog input to the output of a
    Digital-to-Analog Converter (:term:`DAC`), which is initialized with the
    :term:`MSB` high and all other bits low.
    If the analog sample is higher than the :term:`DAC` output the :term:`MSB` keeps high,
    while if the :term:`DAC` output is higher the :term:`MSB` is set low.
    This comparison is performed for each bit from the :term:`MSB` to the :term:`LSB`
    following the same procedure.

    .. figure:: _static/sar_architecture.png
        :scale: 70%
        :alt: Simplified N-bit SAR ADC architecture.
        :align: center

        Simplified N-bit :term:`SAR` :term:`ADC` architecture
        (see https://www.maximintegrated.com/en/app-notes/index.mvp/id/1080).

    Considering a 8-bits :term:`SAR` :term:`ADC`, for the first comparison the
    8-bit register is set to midscale (`10000000`) forcing the :term:`DAC` output
    (:math:`V_{DAC}`) to be half of the reference voltage:
    :math:`V_{DAC}=\frac{V_{REF}}{2}`.
    For the second comparison, if :math:`V_{DAC} < V_{IN}` then the 8-bit register is set
    to `11000000`, if :math:`V_{DAC} > V_{IN}` then the 8-bit register is set to `01000000`.
    In both cases :math:`V_{DAC} = \frac{V_{REF}}{4}`.

    It is essential that :math:`V_{REF}` remains stable through all comparisons
    to be able to perform an accurate conversion from the analog inputs of
    the instrument.

    A potential sources of digitization issues can be modeled by adding
    voltage fluctuations to :math:`V_{REF}` at each stage of the process.

    These fluctuations can be represented by the vectors :math:`\mathit{strength}` and :math:`\mathit{noise}`.
    For a 8-bits :term:`SAR` :term:`ADC`, we have
    :math:`\mathit{strength} = \begin{pmatrix}\mathit{strength}_\mathit{bit7}\\ \vdots\\
    \mathit{strength}_\mathit{bit0}\end{pmatrix}`
    and
    :math:`noise = \begin{pmatrix}\mathit{noise}_\mathit{bit7}\\ \vdots\\ \mathit{noise}_\mathit{bit0}\end{pmatrix}`.

    The fluctuations at each stage of the process (in this case 8 stages) are represented by
    the vector :math:`V^\mathit{perturbated}_\mathit{REF}` with

    .. math::
       \begin{matrix}
       V^\mathit{perturbated}_{\mathit{REF}_\mathit{bit7}} = V_\mathit{REF}
       \left(1 + \mathit{np.random.normal}(\mathit{strength}_\mathit{bit7}, \mathit{noise}_\mathit{bit7})\right)\\
       \cdots\\
       V^\mathit{perturbated}_{\mathit{REF}_\mathit{bit0}} = V_\mathit{REF}
       \left(1 + \mathit{np.random.normal}(\mathit{strength}_\mathit{bit0}, \mathit{noise}_\mathit{bit0})\right)
       \end{matrix}
    """
    min_volt, max_volt = detector.characteristics.adc_voltage_range
    adc_bits = detector.characteristics.adc_bit_resolution

    if len(strengths) != detector.characteristics.adc_bit_resolution:
        raise ValueError(
            f"Expecting a sequence of {adc_bits} elements for parameter 'strengths'."
        )

    if len(noises) != detector.characteristics.adc_bit_resolution:
        raise ValueError(
            f"Expecting a sequence of {adc_bits} elements for parameter 'noises'."
        )

    image_2d: np.ndarray = apply_sar_adc_with_noise(
        signal_2d=detector.signal.array,
        num_rows=detector.geometry.row,
        num_cols=detector.geometry.col,
        strengths=np.asarray(strengths, dtype=float),
        noises=np.asarray(noises, dtype=float),
        max_volt=max_volt,
        adc_bits=adc_bits,
    )

    detector.image.array = image_2d



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


import pytest

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Environment,
    ReadoutProperties,
)


@pytest.fixture
def environment() -> Environment:
    return Environment()


@pytest.fixture
def ccd_geometry() -> CCDGeometry:
    return CCDGeometry(row=1, col=1)


@pytest.fixture
def ccd_characteristics() -> Characteristics:
    return Characteristics()


@pytest.fixture
def ccd_empty(ccd_geometry, environment, ccd_characteristics) -> CCD:
    detector = CCD(
        geometry=ccd_geometry,
        environment=environment,
        characteristics=ccd_characteristics,
    )
    return detector


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(temperature=200.0),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from pathlib import Path
from shutil import copy2

import pytest

from pyxel.configuration import save


@pytest.fixture
def valid_filename(request: pytest.FixtureRequest) -> Path:
    """Get a valid existing YAML filename."""
    filename: Path = request.path / "../data/exposure1.yaml"
    return filename.resolve(strict=True)


@pytest.mark.parametrize(
    "filename",
    [
        "my_config.yaml",
        "other_config.yml",
        "test.YAML",
        "another.YML",
    ],
)
def test_save_with_yaml(filename, tmp_path: Path, valid_filename: Path):
    """Test function 'save'."""
    input_folder: Path = tmp_path / "input"
    input_folder.mkdir()

    # Create an input YAML file
    input_filename: Path = input_folder / filename
    copy2(src=valid_filename, dst=input_filename)
    assert input_filename.exists()

    # Create an output folder
    output_folder: Path = tmp_path / "output"
    output_folder.mkdir()

    # Test
    filename_copied = save(input_filename=input_filename, output_dir=output_folder)
    exp_filename_copied = tmp_path / "output" / filename

    assert filename_copied.exists()
    assert filename_copied == exp_filename_copied
